# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Debugger hook."""

import argparse

from mindinsight.conf import settings
from mindinsight.utils.hook import BaseHook


def enable_debugger_string(string):
    """Convert str to bool"""
    if string.lower() in ('false', '0'):
        return False
    if string.lower() in ('true', '1'):
        return True
    raise ValueError


class EnableDebuggerAction(argparse.Action):
    """Enable debugger action class definition."""

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        enable_debugger = values
        setattr(namespace, self.dest, enable_debugger)


class PortAction(argparse.Action):
    """Port action class definition."""

    MIN_PORT = 1
    MAX_PORT = 65535

    OPEN_PORT_LIMIT = 1024

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        port = values
        if not self.MIN_PORT <= port <= self.MAX_PORT:
            parser.error(f'{option_string} should be chosen from {self.MIN_PORT} to {self.MAX_PORT}')

        setattr(namespace, self.dest, port)


class SessionNumAction(argparse.Action):
    """Session number action class definition."""

    MIN_NUM = 1
    MAX_NUM = 2

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        sesson_num = values
        if not self.MIN_NUM <= sesson_num <= self.MAX_NUM:
            parser.error(f'{option_string} should be chosen from {self.MIN_NUM} to {self.MAX_NUM}')

        setattr(namespace, self.dest, sesson_num)


class MemLimitAction(argparse.Action):
    """Memory limit action class definition."""

    # The unit is MB
    MIN_VALUE = 6 * 1024
    # The limit for int32.
    MAX_VALUE = 2147483647

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        mem_limit = values
        if not self.MIN_VALUE <= mem_limit <= self.MAX_VALUE:
            parser.error(f'{option_string} should be chosen from {self.MIN_VALUE} to {self.MAX_VALUE}')

        setattr(namespace, self.dest, mem_limit)


class MaxGraphNodeSizeAction(argparse.Action):
    """Graph Node limit action class definition."""
    MIN_VALUE = 1
    # The limit for 2000000.
    MAX_VALUE = 2000000

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        max_graph_node_size = values
        if not self.MIN_VALUE <= max_graph_node_size <= self.MAX_VALUE:
            parser.error(f'{option_string} should be chosen from {self.MIN_VALUE} to {self.MAX_VALUE}, default is '
                         f'{settings.MAX_GRAPH_NODE_SIZE}')
        setattr(namespace, self.dest, max_graph_node_size)


class Hook(BaseHook):
    """Hook class definition."""

    def register_startup_arguments(self, parser):
        """
        Hook function to register startup arguments.

        Args:
            parser (ArgumentParser): Specify parser to which arguments are added.
        """
        parser.add_argument(
            '--enable-debugger',
            type=enable_debugger_string,
            action=EnableDebuggerAction,
            default=False,
            help="""
                Enable debugger or not. The value can be True/False/1/0 (case insensitive).
                Default is False.""")

        parser.add_argument(
            '--debugger-port',
            type=int,
            action=PortAction,
            help="""
                Debugger port ranging from %s to %s. Default value is %s.
            """ % (PortAction.MIN_PORT, PortAction.MAX_PORT, settings.DEBUGGER_PORT))

        parser.add_argument(
            '--offline-debugger-mem-limit',
            type=int,
            action=MemLimitAction,
            help="""
                Offline debugger memory limit ranging from %s to %s MB. Default value is %s MB.
            """ % (MemLimitAction.MIN_VALUE, MemLimitAction.MAX_VALUE, settings.OFFLINE_DEBUGGER_MEM_LIMIT))

        parser.add_argument(
            '--max-offline-debugger-session-num',
            type=int,
            action=SessionNumAction,
            help="""
                Max offline debugger session number ranging from %s to %s. Default value is %s.
            """ % (SessionNumAction.MIN_NUM, SessionNumAction.MAX_NUM, settings.MAX_OFFLINE_DEBUGGER_SESSION_NUM))

        parser.add_argument(
            '--max-graph-node-size',
            type=int,
            action=MaxGraphNodeSizeAction,
            help="""
                Offline debugger graph node limit ranging from %s to %s MB. Default value is %s.
            """ % (MaxGraphNodeSizeAction.MIN_VALUE, MaxGraphNodeSizeAction.MAX_VALUE, settings.MAX_GRAPH_NODE_SIZE))



# Copyright 2023 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TensorStatistic for DebuggerTensor."""


class TensorStatistic:
    """The tensor statistic class."""

    def __init__(self, op_type='', op_name='', task_id=0, stream_id=0, time_stamp=0,
                 io='', slot=0, data_size=0, data_type='', shape=(),
                 max_value=float('-inf'), min_value=float('inf'), avg_value=0, count=0,
                 negative_zero_count=0,
                 positive_zero_count=0, nan_count=0, negative_inf_count=0,
                 positive_inf_count=0, zero_count=0):
        self.op_type = op_type
        self.op_name = op_name
        self.task_id = task_id
        self.stream_id = stream_id
        self.time_stamp = time_stamp
        self.io = io
        self.slot = slot
        self.data_size = data_size
        self.data_type = data_type
        self.shape = shape
        self.max_value = max_value
        self.min_value = min_value
        self.avg_value = avg_value
        self.count = count
        self.negative_zero_count = negative_zero_count
        self.positive_zero_count = positive_zero_count
        self.nan_count = nan_count
        self.negative_inf_count = negative_inf_count
        self.positive_inf_count = positive_inf_count
        self.zero_count = zero_count


class SummaryStatistic:
    """The tensor statistic class."""

    def __init__(self, op_type='', op_name='', tensor_name='', inf_number=0, nan_number=0,
                 out_of_range_number=0, total_number=0, none_overflow_number=0, total_iterations=0):
        self.op_type = op_type
        self.op_name = op_name
        self.tensor_name = tensor_name
        self.inf_number = inf_number
        self.nan_number = nan_number
        self.out_of_range_number = out_of_range_number
        self.total_number = total_number
        self.none_overflow_number = none_overflow_number
        self.total_iterations = total_iterations



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""DebuggerServerBase."""
import threading
from abc import abstractmethod
from functools import wraps

from mindinsight.debugger.common.exceptions.exceptions import DebuggerServerRunningError
from mindinsight.debugger.common.log import LOGGER as log


def debugger_server_wrap(func):
    """Wrapper for catch exception."""
    @wraps(func)
    def record_log(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            log.exception(err)
            raise DebuggerServerRunningError(str(err))
    return record_log


class DebuggerServerBase(threading.Thread):
    """
    Debugger Server Base.

    Args:
        cache_store (DebuggerCacheStore): Cache store for debugger server.
        context (DebuggerServerContext): Context for initialize debugger server.
    """

    def __init__(self, cache_store, context):
        super(DebuggerServerBase, self).__init__()
        self._cache_store = cache_store
        self._context = context

    @abstractmethod
    @debugger_server_wrap
    def run(self):
        """Function that should be called when thread started."""

    @abstractmethod
    @debugger_server_wrap
    def stop(self):
        """Stop debugger server."""



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This file is used to define the basic graph.
"""
import time

from enum import Enum
from collections import defaultdict

from mindinsight.datavisual.common.exceptions import NodeNotInGraphError
from mindinsight.datavisual.common.log import logger
from mindinsight.utils.exceptions import ParamMissError
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.domain.graph.base import NodeTypeEnum
from .node import Node


def check_invalid_character(string):
    """Check for invalid characters. These characters will cause frontend crash."""
    invalid_char = {'>', '<', '"'}
    result = set(string).intersection(invalid_char)
    if result:
        raise ParamValueError(f"There are some invalid characters in graph node, invalid string: {string}, "
                              f"unexpected characters: {result}")


class EdgeTypeEnum(Enum):
    """Node edge type enum."""
    CONTROL = 'control'
    DATA = 'data'


class Graph:
    """The `Graph` object is used to describe a graph file."""
    # Limit the size of a single attribute value per node to avoid storing too much data
    MAX_NODE_ATTRIBUTE_VALUE_BYTES = 1024

    # In the same scope, the number of children of the same type exceeds this threshold, and we will combine them.
    MIN_GROUP_NODE_COUNT = 5

    def __init__(self):
        # Used to cache all nodes, and the key is node name, value is `Node` object.
        self._normal_node_map = {}
        self._node_id_map_name = {}

        # The additional caching of Const and Parameter is to handle the Const
        # and Parameter nodes separately later.
        self._const_node_temp_cache = {}
        self._parameter_node_temp_cache = {}

        self._leaf_nodes = {}
        self._full_name_map_name = {}

    def build_graph(self, proto_data):
        """This method is used to build the graph."""
        logger.info("Start to build graph")
        start_time = time.time()

        # Notice:
        # The following methods are interdependent and cannot be switched at will.
        self._parse_data(proto_data)
        self._add_variable_nodes(NodeTypeEnum.PARAMETER.value)
        self._build_aggregation_scope_nodes()
        self._process_independent_layout()
        self._build_name_scope_nodes()

        # Since const nodes are not aggregated, adding them at the end can save a lot of computation.
        self._add_variable_nodes(NodeTypeEnum.CONST.value)
        self._calc_subnode_count()
        self._leaf_nodes = self._get_leaf_nodes()
        self._full_name_map_name = self._get_leaf_node_full_name_map()

        precision = 6
        time_consuming = round(time.time() - start_time, precision)
        logger.info("Build graph end, all node count: %s, const count: %s, parameter count: %s, time-consuming: %s s.",
                    self.normal_node_count, len(self._const_node_temp_cache),
                    len(self._parameter_node_temp_cache), time_consuming)

    def _get_leaf_nodes(self):
        """
        Get all leaf nodes, including normal leaf nodes, const nodes and param nodes.
        """
        leaf_nodes = {}
        for node_name, node in self._normal_node_map.items():
            # update full name
            if not node.full_name:
                node.full_name = node.name
            if not node.type or node.type.endswith('_scope'):
                continue
            leaf_nodes[node_name] = node

        return leaf_nodes

    def _get_leaf_node_full_name_map(self):
        """Get node by debugger name."""
        full_name_map = {}
        for name, node in self._leaf_nodes.items():
            if not node.full_name:
                logger.warning("Node %s does not have full name.", name)
                continue
            full_name_map[node.full_name] = name

        return full_name_map

    def exist_node(self, name):
        """
        Check node exist in graph.

        Args:
            name (str): The node name.

        Returns:
            bool, if node exists, will return True.

        """
        if name is None:
            return False
        return self._is_node_exist(node_name=name)

    def list_node_by_scope(self, scope=None):
        """
        List nodes by the scope of nodes. The scope of a node is the same as its parent node name.

        Args:
            scope (str): A scope of nodes.

        Returns:
            list[dict], a list object contain `Node` object.
        """
        scope = "" if scope is None else scope
        nodes = []
        for node in self._normal_node_map.values():
            if node.scope == scope:
                nodes.append(node.to_dict())
        return nodes

    def search_single_node(self, node_name):
        """
        Search node, and return every layer nodes until this node.

        Args:
            node_name (str): The name of node.

        Returns:
            dict, a dict object, format is :
                item_object = {'nodes': [<Node object>],
                               'scope_name': '<Node scope>',
                               'children': {<item_object>}}
        """
        if node_name and not self.exist_node(name=node_name):
            raise NodeNotInGraphError(node_name=node_name)

        response = {}
        nodes = self.list_node_by_scope()
        response.update({
            'nodes': nodes,
            'scope_name': '',
            'children': {}
        })

        children = response['children']

        index = node_name.find('/')
        while index != -1:
            scope = node_name[:index]
            nodes = self.list_node_by_scope(scope)
            children.update({
                'nodes': nodes,
                'scope_name': scope,
                'children': {}
            })
            children = children['children']

            index = node_name.find('/', index+1)

        return response

    def _parse_data(self, proto_data):
        """
        This method will parse the data and create basic nodes to store in the cache.

        The graph is then built based on the cache.
        """
        raise NotImplementedError("Before you can build a graph, you need to parse the data.")

    def _build_name_scope_nodes(self):
        """
        Build name scope node by every node name.

        We create the name scope node by the slash('/') in the node name.
        For example, if a node name is "Default/add", we generate a scope named 'Default' based on slash('/') and
        create a name scope node named 'Default'.
        """
        logger.info("Start to build name scope nodes.")
        scope_node_map = {}
        for name, node in self._normal_node_map.items():
            index = name.find('/')
            pre_index = None
            while index > 0:
                scope = name[:index]
                scope_node = scope_node_map.get(scope)
                if scope_node is None:
                    if self._is_node_exist(node_name=scope):
                        exist_node = self._get_normal_node(node_name=scope)
                        if exist_node.type == NodeTypeEnum.AGGREGATION_SCOPE.value:
                            # This scope is aggregation scope, so we don't have to do anything.
                            pre_index = index
                            index = name.find('/', pre_index + 1)
                            continue

                        # We find a node name that conflicts with the current scope and rename the node
                        self._update_conflict_node(conflict_name=scope)

                    # We create a node for current scope.
                    scope_node = Node(scope, node_id=scope)
                    scope_node.type = NodeTypeEnum.NAME_SCOPE.value
                    scope_node.scope = '' if pre_index is None else name[:pre_index]
                    scope_node_map.update({scope_node.name: scope_node})

                # Inherit input and output from sub nodes.
                self._inherit_input_output_from_subnode(scope_node, subnode_list=[node])

                pre_index = index
                index = name.find('/', pre_index+1)

        # Cache all the scope node to normal node dict
        for node in scope_node_map.values():
            self._cache_node(node)

    def _update_conflict_node(self, conflict_name):
        conflict_node = self._get_normal_node(node_name=conflict_name)
        base_name = conflict_name.split('/')[-1]
        new_name = Node.create_node_name(scope=conflict_node.scope, base_name=f'({base_name})')
        self._update_node_name_of_cache(conflict_node, new_name, update_parent=True)

    def _inherit_input_output_from_subnode(self, parent_node, subnode_list, filtered_type=None):
        """
        Adds the input and output of all direct child nodes to the current node.

        Args:
            parent_node (Node): The nodes that inherit the input and output of the child nodes.
            subnode_list (list[Node]): A list of child nodes that are inherited from the input and output.
            filtered_type (set(str)): Filter some input and output that do not require inheritance
                                      based on the node type. Default is filter const node.

        Note:
            - Only the inputs and outputs of the external scope are inherited.
            - Before add_const_node method, if the input is a const,
              the scope of the const node is not startswith the name of parent node.
              So in this scenario, we need to filter the const nodes.
        """
        filtered_type = {NodeTypeEnum.CONST.value} if filtered_type is None else filtered_type
        for method in ['inputs', 'outputs', 'proxy_inputs', 'proxy_outputs']:
            for node in subnode_list:
                for item_name, item_attr in getattr(node, method).items():
                    target_node = self._get_normal_node(node_name=item_name)
                    if item_name.startswith(f'{parent_node.name}/'):
                        # Own scope, ignore
                        continue

                    if target_node.type in filtered_type:
                        continue

                    getattr(parent_node, f'add_{method}')(item_name, item_attr)

    def _build_aggregation_scope_nodes(self):
        """
        Under the same scope, the number of nodes of the same type will be aggregated after exceeding the set threshold.

        Note:
            The threshold value refers to the `MIN_GROUP_NODE_COUNT`.
        """
        logger.info("Start to build aggregation scope nodes.")
        group_node_map, filtered_group_names = self._find_group_nodes()

        # create merge scope nodes
        aggregation_scope_node_map = {}
        for i, group_name in enumerate(filtered_group_names):
            slash_index = group_name.rfind('/')
            if slash_index != -1:
                scope, op_type = group_name[:slash_index], group_name[slash_index+1:]
            else:
                scope, op_type = '', group_name

            count = len(group_node_map.get(group_name))
            aggregation_node_name = Node.create_node_name(scope=scope, base_name=f'{op_type}[{count}]_{i}')
            aggregation_scope_node = Node(name=aggregation_node_name, node_id=aggregation_node_name)
            aggregation_scope_node.subnode_count = count
            aggregation_scope_node.scope = scope
            aggregation_scope_node.type = NodeTypeEnum.AGGREGATION_SCOPE.value

            # Update the name and scope of all children nodes
            for node in group_node_map[group_name]:
                base_name = node.name.split('/')[-1]
                new_name = Node.create_node_name(scope=aggregation_node_name, base_name=base_name)
                node.scope = aggregation_node_name

                # Since the name scope has not been created, there is no need to update the parent node.
                self._update_node_name_of_cache(node, new_name, update_parent=False)

            # Cache this node
            self._cache_node(aggregation_scope_node)
            aggregation_scope_node_map.update({group_name: aggregation_scope_node})

        # Adds the input and output of all direct child nodes to the current node.
        for group_name, node in aggregation_scope_node_map.items():
            self._inherit_input_output_from_subnode(node, group_node_map[group_name])

    def _find_group_nodes(self):
        """
        Find nodes that can be grouped into a group.

        For direct child nodes in a scope, we divide them into multiple groups by node type.
        However, we will exclude several types of child nodes,
        because these types of nodes are not operational nodes.
        """
        exclude_types = {
            NodeTypeEnum.CONST.value,
            NodeTypeEnum.NAME_SCOPE.value,
        }

        group_node_map = defaultdict(list)
        for node in self._normal_node_map.values():
            if node.type in exclude_types:
                continue
            group_name = Node.create_node_name(scope=node.scope, base_name=node.type)
            group_node_map[group_name].append(node)

        # filter can group scope.
        filtered_group_names = []
        for name, nodes in group_node_map.items():
            if len(nodes) < self.MIN_GROUP_NODE_COUNT:
                continue
            filtered_group_names.append(name)

        return group_node_map, filtered_group_names

    def _add_variable_nodes(self, node_type):
        """
        We create the Const nodes or Parameter nodes in this method.

        Args:
            node_type (str): Decide which type of node to add.
                             Optional is `NodeTypeEnum.CONST.value` and `NodeTypeEnum.PARAMETER.value`.

        Note:
            This method relies on the presence of data in the const cache or parameter cache.
        """
        logger.info("Start to add %s nodes to each scope in graph.", node_type)
        node_map = {}
        for node in self._normal_node_map.values():
            for src_name, input_attr in dict(node.inputs).items():

                if node_type == NodeTypeEnum.CONST.value and not self._const_node_temp_cache.get(src_name):
                    continue

                if node_type == NodeTypeEnum.PARAMETER.value and not self._parameter_node_temp_cache.get(src_name):
                    continue

                variable_name = Node.create_node_name(scope=node.scope, base_name=src_name)
                if node_map.get(variable_name):
                    # There is no need to create the node repeatedly
                    variable_node = node_map.get(variable_name)
                else:
                    cache_node = self._get_normal_node(node_name=src_name)
                    variable_node = Node(name=variable_name, node_id=variable_name)
                    Node.copy_node_without_input_output(cache_node, variable_node)
                    variable_node.scope = node.scope

                variable_node.add_outputs(dst_name=node.name, output_attr=dict(input_attr))
                node_map.update({variable_name: variable_node})

                node.delete_inputs(src_name)
                node.add_inputs(variable_name, dict(input_attr))

        for node in node_map.values():
            self._cache_node(node)

        # Remove nodes that are not used in the cache.
        if node_type == NodeTypeEnum.CONST.value:
            unused_names = set(self._const_node_temp_cache) - set(node_map)
        elif node_type == NodeTypeEnum.PARAMETER.value:
            unused_names = set(self._parameter_node_temp_cache) - set(node_map)
        else:
            raise ParamValueError("The node type should be const or parameter.")

        self._delete_nodes_of_cache(unused_names)

    def _calc_subnode_count(self):
        """Calc all the direct sub node count."""
        subnode_count_map = defaultdict(int)
        for node in self._normal_node_map.values():
            if not node.scope:
                continue

            if not self._is_node_exist(node_name=node.scope):
                logger.warning("Can not find a scope node by the given name(%s), "
                               "the name scope nodes may not have been created.", node.scope)
                continue
            subnode_count_map[node.scope] = subnode_count_map[node.scope] + 1

        for name, count in subnode_count_map.items():
            node = self._get_normal_node(node_name=name)
            node.subnode_count = count

    def _get_normal_node(self, node_id=None, node_name=None):
        """Query node by node id or node name."""
        if node_id is not None:
            name = self._node_id_map_name.get(node_id)
            node = self._normal_node_map.get(name)
            return node

        if node_name is not None:
            return self._normal_node_map.get(node_name)

        raise ParamMissError('Method requires an argument that is not None.')

    def _is_node_exist(self, node_id=None, node_name=None):
        """Check node is exist."""
        if node_id is not None:
            return bool(self._node_id_map_name.get(node_id))

        if node_name is not None:
            return bool(self._normal_node_map.get(node_name))

        raise ParamMissError('Method requires an argument that is not None.')

    @property
    def normal_node_count(self):
        """Get the normal node count."""
        return len(self._normal_node_map)

    def _cache_node(self, node):
        """Store the node in the cache."""
        # Notice:
        # The additional caching of Const and Parameter is to handle the Const and Parameter nodes separately later.
        if node.type == NodeTypeEnum.CONST.value:
            self._const_node_temp_cache.update({node.name: node})
        if node.type == NodeTypeEnum.PARAMETER.value:
            self._parameter_node_temp_cache.update({node.name: node})

        self._normal_node_map.update({node.name: node})
        self._node_id_map_name.update({node.node_id: node.name})

    def _delete_nodes_of_cache(self, node_names):
        """Delete node from cache."""
        logger.debug("These nodes will be removed from the cache, node names: %s.", str(node_names))
        for name in node_names:

            if self._parameter_node_temp_cache.get(name):
                self._parameter_node_temp_cache.pop(name)
            if self._const_node_temp_cache.get(name):
                self._const_node_temp_cache.pop(name)

            node = self._get_normal_node(node_name=name)
            self._normal_node_map.pop(name)
            if self._node_id_map_name.get(node.node_id):
                self._node_id_map_name.pop(node.node_id)

    def _update_node_name_of_cache(self, node, new_name, update_parent=False):
        """
        Update a node name which is stored in cache.

        Args:
            node (Node): The node that will be renamed.
            new_name (str): The new name.
            update_parent (bool): Determines whether the input and output of the parent node need to be updated.
        """
        logger.debug('Update node name of cache, node(%s), new name is %s.', str(node), new_name)
        origin_name = node.name
        node.name = new_name
        node.scope = new_name.rsplit('/', 1)[0]

        # Find all nodes that need to modify the input and input
        update_node_map = {}
        for method in ['inputs', 'outputs', 'proxy_inputs', 'proxy_outputs']:
            for target_name in getattr(node, method):
                target_node = self._get_normal_node(node_name=target_name)
                if target_node is None:
                    message = f"Node should not be None, name: {target_name}, {method}: {list(getattr(node, method))}."
                    logger.error(message)
                    continue

                update_node_map.update({target_name: target_node})

                if not update_parent:
                    continue

                slash_index = target_name.find('/')
                while slash_index != -1:
                    scope_name = target_name[:slash_index]
                    slash_index = target_name.find('/', slash_index+1)

                    if update_node_map.get(scope_name):
                        continue

                    scope_node = self._get_normal_node(node_name=scope_name)
                    if scope_node is None:
                        message = f"Can not find the scope node by scope name({scope_name}), " \
                                  f"may be this scope node has not been built."
                        logger.debug(message)
                        continue

                    update_node_map.update({scope_name: scope_node})

        # Update the input and output of the nodes
        for target_node in update_node_map.values():
            for method in ['inputs', 'outputs', 'proxy_inputs', 'proxy_outputs']:
                attr_temp = getattr(target_node, method).get(origin_name)
                if attr_temp is None:
                    # This method does not have this node, so it is skipped
                    continue

                # Delete the old attribute and update new name to source node or destination node.
                getattr(target_node, f'delete_{method}')(origin_name)
                getattr(target_node, f'add_{method}')(new_name, attr_temp)

        # Delete the origin node in cache.
        self._delete_nodes_of_cache(node_names=[origin_name])
        self._cache_node(node)

    def _process_independent_layout(self):
        """Handle separate layout nodes."""
        independent_layout_node_map = {}
        for node in self._normal_node_map.values():
            base_name = node.name.split('/')[-1]
            if node.type == NodeTypeEnum.AGGREGATION_SCOPE.value and NodeTypeEnum.PARAMETER.value in base_name:
                independent_layout_node_map[node.name] = node

        # Find all sub nodes
        subnode_map = defaultdict(list)
        for node in self._normal_node_map.values():
            if independent_layout_node_map.get(node.scope):
                subnode_map[node.scope].append(node)

        # Notice:
        # The following processing is only done for the parameter node, other types of nodes are not processed.
        # Later, when you need to extend to other nodes, the code needs to be adjusted.
        for scope_node in independent_layout_node_map.values():
            scope_node.independent_layout = True

            method = 'outputs'
            for target_name, target_attr in dict(getattr(scope_node, method)).items():
                proxy_attr = dict(edge_type=target_attr['edge_type'])

                target_node = self._get_normal_node(node_name=target_name)
                getattr(target_node, 'add_proxy_inputs')(scope_node.name, proxy_attr)

                # Note:
                # If the source node and the destination node are not in the same scope,
                # the proxy node is presented as scope in order to simplify the flow of the display data.
                # For example, the data flow is parameter[5]_1 -> add[5]_1/add1
                # we create a scope proxy node(add[5]_1) for parameter[5]_1,
                # so there is a proxy data flow parameter[5]_1 -> add[5]_1 instead of parameter[5]_1 -> add[5]_1/add1.
                if target_node.scope == scope_node.scope:
                    getattr(scope_node, f'add_proxy_{method}')(target_name, proxy_attr)
                else:
                    target_scope_node = self._get_normal_node(node_name=target_node.scope)
                    getattr(scope_node, f'add_proxy_{method}')(target_node.scope, proxy_attr)
                    getattr(target_scope_node, 'add_proxy_inputs')(scope_node.name, proxy_attr)

            for subnode in subnode_map[scope_node.name]:
                subnode.independent_layout = True
                for target_name, target_attr in dict(getattr(subnode, method)).items():
                    proxy_attr = dict(edge_type=target_attr['edge_type'])
                    target_node = self._get_normal_node(node_name=target_name)
                    if target_node.scope == scope_node.scope:
                        getattr(subnode, f'add_proxy_{method}')(target_name, proxy_attr)
                    else:
                        getattr(subnode, f'add_proxy_{method}')(target_node.scope, proxy_attr)

                    input_attr = getattr(target_node, 'inputs')[subnode.name]
                    input_attr['independent_layout'] = True
                    target_node.add_inputs(subnode.name, input_attr)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Parse utils module."""

import os
import stat
from functools import cmp_to_key

import xlsxwriter

from mindinsight.domain.graph.base import InputType, OutputType


class Toolkit:
    """Toolkit."""

    PLACEHOLDER = '-'

    def __init__(self, dump_dir, constants, parameters, operators):
        self.dump_dir = dump_dir
        self.constants = constants
        self.parameters = parameters
        self.operators = operators

    def export_xlsx(self, file_path):
        """
        Export graph data to Excel file.

        Args:
            file_path (str) : Excel file path.
        """
        file_path = os.path.realpath(file_path)
        target_dir = os.path.dirname(file_path)
        if not os.path.isdir(target_dir):
            print(f'Directory {target_dir} not exists')
            return

        workbook = xlsxwriter.Workbook(file_path)

        # text_v_align: 1-top, 2-middle, 3-bottom
        # text_h_align: 1-left, 2-center, 3-right
        styles = dict(
            header_left_fmt=workbook.add_format(dict(
                text_v_align=2, text_h_align=1,
                font_color='#000000', bg_color='#d9d9d9',
                bold=True,
            )),
            header_center_fmt=workbook.add_format(dict(
                text_v_align=2, text_h_align=2,
                font_color='#000000', bg_color='#d9d9d9',
                bold=True,
            )),
            content_left_fmt=workbook.add_format(dict(
                text_v_align=2, text_h_align=1,
                text_wrap=False,
            )),
            content_center_fmt=workbook.add_format(dict(
                text_v_align=2, text_h_align=2,
                text_wrap=False,
            )),
            content_wrapped_fmt=workbook.add_format(dict(
                text_v_align=2, text_h_align=1,
                text_wrap=True,
            )),
        )

        self._add_info_worksheet(workbook, styles)
        self._add_constant_worksheet(workbook, styles)
        self._add_parameter_worksheet(workbook, styles)
        self._add_operator_worksheet(workbook, styles)
        self._add_statistics_worksheet(workbook, styles)
        self._add_source_worksheet(workbook, styles)

        for worksheet in workbook.sheetnames.values():
            worksheet.freeze_panes(1, 0)
            worksheet.freeze_panes(1, 1)

        workbook.close()
        os.chmod(file_path, stat.S_IRUSR)

    def _write_columns(self, worksheet, metas):
        """
        Write columns according to column metas.

        Args:
            worksheet (Workbook): Target worksheet.
            metas (list): Column metas.
        """
        for index, (column, fmt, width) in enumerate(metas):
            worksheet.set_column(index, index, width)
            worksheet.write(0, index, column, fmt)
        worksheet.autofilter(0, 0, 0, len(metas) - 1)

    def _convert_column_indices(self, metas):
        """
        Convert column metas into indices mapping.

        Args:
            metas (list): Column metas.

        Returns:
            dict, holds the indicess of columns.
        """
        mapping = {}
        for index, (name, _, _) in enumerate(metas):
            mapping[name] = index
        return mapping

    def _add_info_worksheet(self, workbook, styles):
        """
        Add info worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        worksheet = workbook.add_worksheet('info')

        # column metas contain column names, styles and widths
        column_metas = [
            ('argument', styles['header_center_fmt'], 20),
            ('value', styles['header_left_fmt'], 150),
        ]
        self._write_columns(worksheet, column_metas)

        indices = self._convert_column_indices(column_metas)
        worksheet.write(1, indices.get('argument'), 'dump-dir', styles['content_center_fmt'])
        worksheet.write(1, indices.get('value'), self.dump_dir or '', styles['content_left_fmt'])

    def _get_dtype_content(self, dtype):
        """
        Get dtype content.

        Args:
            dtype (Any): Tensor dtype.

        Returns:
            str, dtype content.
        """
        return Toolkit.PLACEHOLDER if dtype is None else str(dtype)

    def _get_shape_content(self, shape):
        """
        Get shape content.

        Args:
            shape (Any): Tensor shape.

        Returns:
            str, shape content.
        """
        return Toolkit.PLACEHOLDER if shape is None else str(shape)

    def _get_operator_input_content(self, operator, input_types):
        """
        Add operator input content.

        Args:
            operator (Operator): Operator.
            input_types (dict): Input types.

        Returns:
            tuple, input content, input dtype content and input shape content.
        """
        input_content = ''
        input_dtype_content = ''
        input_shape_content = ''

        for op_input in operator.inputs:
            if op_input.type == InputType.OPERATOR:
                op = input_types[InputType.OPERATOR][op_input.op_id]
                input_content += f'{op.type}_{op.name}' if op.type == 'Load' else f'{op.type}_{op.op_id}'
                if op_input.info:
                    input_dtype_content += self._get_dtype_content(op_input.info['dtype'])
                    input_shape_content += self._get_dtype_content(op_input.info.get('shape'))
                else:
                    input_dtype_content += Toolkit.PLACEHOLDER
                    input_shape_content += Toolkit.PLACEHOLDER
            elif op_input.type == InputType.PARAMETER:
                input_content += op_input.name
                param = input_types[InputType.PARAMETER][op_input.name]
                if param.output:
                    input_dtype_content += self._get_dtype_content(param.output.info['dtype'])
                    input_shape_content += self._get_shape_content(param.output.info.get('shape'))
                else:
                    input_dtype_content += Toolkit.PLACEHOLDER
                    input_shape_content += Toolkit.PLACEHOLDER
            elif op_input.type == InputType.CONSTANT:
                input_content += op_input.name
                cst = input_types[InputType.CONSTANT][op_input.name]
                if cst.output.type == OutputType.TENSOR:
                    input_dtype_content += self._get_dtype_content(cst.output.info.get('dtype'))
                    input_shape_content += self._get_shape_content(cst.output.info.get('shape'))
                else:
                    input_dtype_content += Toolkit.PLACEHOLDER
                    input_shape_content += Toolkit.PLACEHOLDER
            else:
                input_content += op_input.name
                input_dtype_content += Toolkit.PLACEHOLDER
                input_shape_content += Toolkit.PLACEHOLDER

            input_content += '\n'
            input_dtype_content += '\n'
            input_shape_content += '\n'

        return input_content.strip(), input_dtype_content.strip(), input_shape_content.strip()

    def _get_operator_type_content(self, operator):
        """
        Get operator downstream content.

        Args:
            operator (Operator): Operator.
            mapping (dict): Operator mapping.

        Returns:
            str, downstream content.
        """
        if operator.type == 'Load':
            return f'{operator.type}_{operator.name}'
        return f'{operator.type}_{operator.op_id}'

    def _get_operator_output_content(self, operator):
        """
        Get operator output content.

        Args:
            operator (Operator): Operator.

        Returns:
            tuple, output dtype content and output shape content.
        """
        output_dtype_content = ''
        output_shape_content = ''
        if operator.output:
            if operator.output.type == OutputType.TENSOR:
                output_dtype_content = operator.output.info['dtype']
                output_shape_content = str(operator.output.info['shape'])
            elif operator.output.type == OutputType.TUPLE:
                output_dtype_content = '\n'.join([
                    Toolkit.PLACEHOLDER if dtype is None else dtype
                    for dtype in operator.output.info['dtypes']
                ])
                output_shape_content = '\n'.join([
                    Toolkit.PLACEHOLDER if shape is None else str(shape)
                    for shape in operator.output.info['shapes']
                ])
        return output_dtype_content, output_shape_content

    def _get_operator_downstream_content(self, operator, mapping):
        """
        Get operator downstream content.

        Args:
            operator (Operator): Operator.
            mapping (dict): Operator mapping.

        Returns:
            str, downstream content.
        """
        content = ''
        for op_id in operator.downstream:
            op = mapping[op_id]
            content += f'{op.type}_{op.op_id}' + '\n'
        return content.strip()

    def _get_operator_stack_content(self, operator):
        """
        Get operator stack content.

        Args:
            operator (Operator): Operator.

        Returns:
            str, stack content.
        """
        content = '\n'.join([str(source) for source in operator.stack])
        return content.strip()

    def _add_operator_worksheet(self, workbook, styles):
        """
        Add operator worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        operator_mapping = dict((operator.op_id, operator) for operator in self.operators)
        input_types = {
            InputType.CONSTANT: dict((constant.name, constant) for constant in self.constants),
            InputType.PARAMETER: dict((parameter.name, parameter) for parameter in self.parameters),
            InputType.OPERATOR: operator_mapping,
        }

        worksheet = workbook.add_worksheet('operator')

        # column metas contain column names, styles and widths
        column_metas = [
            ('operator', styles['header_left_fmt'], 30),
            ('input', styles['header_left_fmt'], 30),
            ('input_dtype', styles['header_left_fmt'], 20),
            ('input_shape', styles['header_left_fmt'], 25),
            ('output_dtype', styles['header_left_fmt'], 20),
            ('output_shape', styles['header_left_fmt'], 25),
            ('downstream', styles['header_left_fmt'], 30),
            ('name', styles['header_center_fmt'], 10),
            ('attrs', styles['header_left_fmt'], 30),
            ('scope', styles['header_left_fmt'], 20),
            ('full_name', styles['header_left_fmt'], 20),
            ('rank_id', styles['header_left_fmt'], 20),
            ('graph_name', styles['header_left_fmt'], 30),
            ('stack', styles['header_left_fmt'], 150),
        ]
        self._write_columns(worksheet, column_metas)

        indices = self._convert_column_indices(column_metas)
        for index, operator in enumerate(self.operators):
            operator_type_content = self._get_operator_type_content(operator)
            worksheet.write(index + 1, indices.get('operator'), operator_type_content, styles['content_left_fmt'])

            if operator.type == 'make_tuple':
                worksheet.write(index + 1, indices.get('rank_id'), operator.rank_id, styles['content_left_fmt'])
                worksheet.write(index + 1, indices.get('graph_name'), operator.graph_name, styles['content_left_fmt'])
                continue

            input_content, input_dtype_content, input_shape_content = \
                self._get_operator_input_content(operator, input_types)
            worksheet.write(index + 1, indices.get('input'), input_content, styles['content_wrapped_fmt'])
            worksheet.write(index + 1, indices.get('input_dtype'), input_dtype_content, styles['content_wrapped_fmt'])
            worksheet.write(index + 1, indices.get('input_shape'), input_shape_content, styles['content_wrapped_fmt'])

            output_dtype_content, output_shape_content = self._get_operator_output_content(operator)
            worksheet.write(index + 1, indices.get('output_dtype'), output_dtype_content, styles['content_wrapped_fmt'])
            worksheet.write(index + 1, indices.get('output_shape'), output_shape_content, styles['content_wrapped_fmt'])

            downstream_content = self._get_operator_downstream_content(operator, operator_mapping)
            worksheet.write(index + 1, indices.get('downstream'), downstream_content, styles['content_wrapped_fmt'])

            worksheet.write(index + 1, indices.get('name'), operator.name, styles['content_center_fmt'])
            worksheet.write(index + 1, indices.get('attrs'), str(operator.attrs), styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('scope'), operator.scope, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('full_name'), operator.full_name, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('rank_id'), operator.rank_id, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('graph_name'), operator.graph_name, styles['content_left_fmt'])

            stack_content = self._get_operator_stack_content(operator)
            worksheet.write(index + 1, indices.get('stack'), stack_content, styles['content_wrapped_fmt'])

    def _add_parameter_worksheet(self, workbook, styles):
        """
        Add parameter worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        worksheet = workbook.add_worksheet('parameter')

        # column metas contain column names, styles and widths
        column_metas = [
            ('name', styles['header_left_fmt'], 50),
            ('output_dtype', styles['header_left_fmt'], 20),
            ('output_shape', styles['header_left_fmt'], 25),
            ('downstream', styles['header_left_fmt'], 30),
            ('rank_id', styles['header_left_fmt'], 20),
            ('graph_name', styles['header_left_fmt'], 30),
        ]
        self._write_columns(worksheet, column_metas)

        indices = self._convert_column_indices(column_metas)
        operator_mapping = dict((operator.op_id, operator) for operator in self.operators)
        for index, parameter in enumerate(self.parameters):
            worksheet.write(index + 1, indices.get('name'), parameter.name, styles['content_left_fmt'])
            worksheet.write(
                index + 1, indices.get('output_dtype'),
                parameter.output.info['dtype'], styles['content_left_fmt'])
            worksheet.write(
                index + 1, indices.get('output_shape'),
                str(parameter.output.info['shape']), styles['content_left_fmt'])

            downstream_nodes = [operator_mapping[op_id] for op_id in parameter.downstream]
            downstream_content = ''
            for op in downstream_nodes:
                # Load is virtual operator for network parameters.
                # It is recommended to export its name rather than its op_id.
                if op.type == 'Load':
                    downstream_content += f'{op.type}_{op.name}' + '\n'
                else:
                    downstream_content += f'{op.type}_{op.op_id}' + '\n'
            worksheet.write(
                index + 1, indices.get('downstream'),
                downstream_content.strip(), styles['content_wrapped_fmt'])

            worksheet.write(index + 1, indices.get('rank_id'), parameter.rank_id, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('graph_name'), parameter.graph_name, styles['content_left_fmt'])

    def _add_constant_worksheet(self, workbook, styles):
        """
        Add constant worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        worksheet = workbook.add_worksheet('constant')

        # column metas contain column names, styles and widths
        column_metas = [
            ('name', styles['header_left_fmt'], 10),
            ('value', styles['header_left_fmt'], 30),
            ('downstream', styles['header_left_fmt'], 30),
            ('rank_id', styles['header_left_fmt'], 20),
            ('graph_name', styles['header_left_fmt'], 30),
        ]
        self._write_columns(worksheet, column_metas)

        indices = self._convert_column_indices(column_metas)
        operator_mapping = dict((operator.op_id, operator) for operator in self.operators)
        for index, constant in enumerate(self.constants):
            worksheet.write(index + 1, indices.get('name'), constant.name, styles['content_left_fmt'])

            if constant.output.type == OutputType.NONE:
                value_content = 'NONE'
            elif constant.output.type == OutputType.TENSOR:
                value_content = 'TENSOR'
            else:
                value_content = constant.output.info['value']
            worksheet.write(index + 1, indices.get('value'), value_content, styles['content_left_fmt'])

            downstream_nodes = [operator_mapping[op_id] for op_id in constant.downstream]
            downstream_content = ''
            for op in downstream_nodes:
                if op.type == 'Load':
                    downstream_content += f'{op.type}_{op.name}' + '\n'
                else:
                    downstream_content += f'{op.type}_{op.op_id}' + '\n'
            worksheet.write(
                index + 1, indices.get('downstream'),
                downstream_content.strip(), styles['content_wrapped_fmt'])

            worksheet.write(index + 1, indices.get('rank_id'), constant.rank_id, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('graph_name'), constant.graph_name, styles['content_left_fmt'])

    def _add_statistics_worksheet(self, workbook, styles):
        """
        Add statistics worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        worksheet = workbook.add_worksheet('statistics')

        # column metas contain column names, styles and widths
        column_metas = [
            ('operator', styles['header_left_fmt'], 30),
            ('count', styles['header_center_fmt'], 20),
        ]
        self._write_columns(worksheet, column_metas)

        operator_type_set = set()
        for operator in self.operators:
            operator_type_set.add(operator.type)

        operator_types = sorted(list(operator_type_set))
        stats = dict(zip(operator_types, [0]*len(operator_types)))
        for operator in self.operators:
            stats[operator.type] += 1

        indices = self._convert_column_indices(column_metas)
        for index, operator_type in enumerate(operator_types):
            worksheet.write(index + 1, indices.get('operator'), operator_type, styles['content_left_fmt'])
            worksheet.write(index + 1, indices.get('count'), stats[operator_type], styles['content_center_fmt'])

    def _add_source_worksheet(self, workbook, styles):
        """
        Add source worksheet.

        Args:
            workbook (WorkBook): Excel workbook.
            styles (dict): Workbook styles.
        """
        worksheet = workbook.add_worksheet('source')

        # column metas contain column names, styles and widths
        column_metas = [
            ('stack', styles['header_left_fmt'], 150),
            ('operator', styles['header_left_fmt'], 30),
            ('full_name', styles['header_left_fmt'], 20),
            ('rank_id', styles['header_left_fmt'], 20),
            ('graph_name', styles['header_left_fmt'], 30),
        ]
        self._write_columns(worksheet, column_metas)

        source_mapping = {}
        for operator in self.operators:
            if not operator.stack:
                continue
            stack = [str(source) for source in operator.stack if source.file_path]
            key = '\n'.join(stack)
            if key in source_mapping:
                source_mapping[key].append(operator)
            else:
                source_mapping[key] = [operator]

        def compare_op(op1, op2):
            if op1.op_id.isdigit() and op2.op_id.isdigit():
                return int(op1.op_id) - int(op2.op_id)
            if op1.op_id < op2.op_id:
                return -1
            if op1.op_id > op2.op_id:
                return 1
            return 0

        row = 0
        indices = self._convert_column_indices(column_metas)
        for key in source_mapping:
            operators = source_mapping[key]
            operators.sort(key=cmp_to_key(compare_op))

            if len(operators) == 1:
                worksheet.write(row + 1, indices.get('stack'), key, styles['content_wrapped_fmt'])
            else:
                worksheet.merge_range(
                    row + 1, indices.get('stack'),
                    row+len(operators), 0, key, styles['content_wrapped_fmt'])

            for index, operator in enumerate(operators):
                operator_content = f'{operator.type}_{operator.op_id}'
                worksheet.write(
                    row + index + 1, indices.get('operator'),
                    operator_content, styles['content_left_fmt'])
                worksheet.write(
                    row + index + 1, indices.get('full_name'),
                    operator.full_name, styles['content_left_fmt'])
                worksheet.write(
                    row + index + 1, indices.get('rank_id'),
                    operator.rank_id, styles['content_left_fmt'])
                worksheet.write(
                    row + index + 1, indices.get('graph_name'),
                    operator.graph_name, styles['content_left_fmt'])

            row += len(operators)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Base loader generator."""
from abc import abstractmethod

MAX_DATA_LOADER_SIZE = 15


class LoaderGenerator:
    """Base loader generator for loader generators."""
    @abstractmethod
    def generate_loaders(self, loader_pool, summaries_info):
        """
        Abstract method for generating loaders.

        Args:
            loader_pool (dict[str, LoaderStruct]): Current loader pool in data_manager.
            summaries_info (list): Summaries info list.

        Returns:
            dict[str, LoaderStruct], a dict of `Loader`.

        """

    @abstractmethod
    def check_train_job_exist(self, train_id):
        """
        Abstract method for checking if train job exists.

        Args:
            train_id (str): Train ID.

        Returns:
            bool, if train job exists, return True.

        """

    @abstractmethod
    def generate_loader_by_train_id(self, train_id):
        """
        Abstract method for generating loader by train id.

        Args:
            train_id (str): Train ID.

        Returns:
            dict[str, LoaderStruct], a dict of `Loader`.

        """



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define custom exception."""
from enum import unique

from mindinsight.utils.constant import WizardErrors
from mindinsight.utils.exceptions import MindInsightException


@unique
class WizardErrorCodes(WizardErrors):
    """Wizard error codes."""
    CODE_SYNTAX_ERROR = 1
    OS_PERMISSION_ERROR = 2
    COMMAND_ERROR = 3
    TEMPLATE_FILE_ERROR = 4


class CodeSyntaxError(MindInsightException):
    """The CodeSyntaxError class definition."""
    def __init__(self, msg):
        super(CodeSyntaxError, self).__init__(WizardErrorCodes.CODE_SYNTAX_ERROR, msg)


class OSPermissionError(MindInsightException):
    def __init__(self, msg):
        super(OSPermissionError, self).__init__(WizardErrorCodes.OS_PERMISSION_ERROR, msg)


class CommandError(MindInsightException):
    def __init__(self, msg):
        super(CommandError, self).__init__(WizardErrorCodes.COMMAND_ERROR, msg)


class TemplateFileError(MindInsightException):
    def __init__(self, msg):
        super(TemplateFileError, self).__init__(WizardErrorCodes.TEMPLATE_FILE_ERROR, msg)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
DataLoader is an adapter for all other loaders.

This module can identify what loader should be used to load data.
"""

from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.data_transform.ms_data_loader import MSDataLoader
from mindinsight.datavisual.common import exceptions


class DataLoader:
    """
    The adapter of all kinds of loaders.

    Args:
        summary_dir (str): A directory path.
    """
    def __init__(self, summary_dir):
        self._summary_dir = summary_dir
        self._loader = None
        self._is_integrity = True

    @property
    def is_integrity(self):
        return self._is_integrity

    @is_integrity.setter
    def is_integrity(self, integrity):
        self._is_integrity = integrity

    def load(self, executor=None):
        """Load the data when loader is exist.

        Args:
            executor (Optional[Executor]): The executor instance.

        Returns:
            bool, True if the loader is finished loading.
        """

        if self._loader is None:
            ms_dataloader = MSDataLoader(self._summary_dir)
            loaders = [ms_dataloader]
            for loader in loaders:
                if loader.filter_valid_files():
                    self._loader = loader
                    break

            if self._loader is None:
                logger.warning("No valid files can be loaded, summary_dir: %s.", self._summary_dir)
                raise exceptions.SummaryLogPathInvalid()
        self.is_integrity = self._loader.is_integrity
        return self._loader.load(executor)

    def get_events_data(self):
        """
        Get events data from log file.

        Returns:
            EventsData, indicates events data.
        """
        return self._loader.get_events_data()

    def has_valid_files(self):
        """
        Check the directory for valid files.

        Returns:
            bool, if the directory has valid files, return True.
        """
        ms_dataloader = MSDataLoader(self._summary_dir)
        return bool(ms_dataloader.filter_valid_files())



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the utils."""
import enum
import gc
import os
import stat
import re
import struct
import tempfile
import time
from functools import wraps

import numpy as np

from mindinsight.conf import settings
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamTypeError, DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.proto.debug_grpc_pb2 import EventReply
from mindinsight.domain.graph.base import NodeTypeEnum
from mindinsight.domain.graph.proto.ms_graph_pb2 import DataType
from mindinsight.utils.tensor import Statistics

# translate the MindSpore type to numpy type.
NUMPY_TYPE_MAP = {
    'DT_BOOL': bool,

    'DT_INT8': np.int8,
    'DT_INT16': np.int16,
    'DT_INT32': np.int32,
    'DT_INT64': np.int64,

    'DT_UINT8': np.uint8,
    'DT_UINT16': np.uint16,
    'DT_UINT32': np.uint32,
    'DT_UINT64': np.uint64,

    'DT_FLOAT16': np.float16,
    'DT_FLOAT32': np.float32,
    'DT_FLOAT64': np.float64,

    'DT_STRING': np.str_,
    'DT_TYPE': np.str_
}

# The buffer is used for the MI side of the offline debugger, the unit is MB.
BUFFER_MS = 1024
# The offline debugger need at least 2g memory space.
MAX_CACHE_SPACE_MB = max(((settings.OFFLINE_DEBUGGER_MEM_LIMIT - BUFFER_MS) // 3), 2 * 1024)
MAX_MS_CACHE_SPACE_MB = settings.OFFLINE_DEBUGGER_MEM_LIMIT - BUFFER_MS - MAX_CACHE_SPACE_MB
MAX_CACHE_SPACE_BYTES = MAX_CACHE_SPACE_MB * 1024 * 1024
# The debugger need to cache 2 tensors.
MAX_SINGLE_TENSOR_CACHE_BYTES = 1024 * 1024 * 1024


@enum.unique
class ReplyStates(enum.Enum):
    """Define the status of reply."""
    SUCCESS = 0
    FAILED = -1


@enum.unique
class ServerStatus(enum.Enum):
    """The status of debugger server."""
    PENDING = 'pending'  # no client session has been connected
    RECEIVE_GRAPH = 'receive_graph'  # the client session has sent the graph
    WAITING = 'waiting'  # the client session is ready
    RUNNING = 'running'  # the client session is running a script
    MISMATCH = 'mismatch'  # the version of MindSpore and MindInsight is not matched
    SENDING = 'sending'  # the request is in cache but not be sent to client
    NODE_TOO_LARGE = 'node_too_large'  # the nodes is too large to analyze

    @classmethod
    def is_normal_state(cls, state):
        """
        Check the server is in normal state.

        Args:
            state (str): The value of ServerStatus.

        Returns:
            bool, if current server is working successfully.
        """
        normal_state = [cls.RECEIVE_GRAPH.value,
                        cls.WAITING.value,
                        cls.RUNNING.value,
                        cls.SENDING.value]
        return state in normal_state


@enum.unique
class Streams(enum.Enum):
    """Define the enable streams to be deal with."""

    COMMAND = "command"
    DATA = "data"
    METADATA = "metadata"
    GRAPH = 'node'
    TENSOR = 'tensor'
    WATCHPOINT = 'watchpoint'
    WATCHPOINT_HIT = 'watchpoint_hit'
    DEVICE = 'device'
    GRAPH_HISTORY = 'graph_history'


class RunLevel(enum.Enum):
    """
    Run Level enum, it depends on whether the program is executed node by node,
    step by step, or in recheck phase.
    """
    NODE = "node"
    STEP = "step"
    RECHECK = "recheck"


class ViewCommandLevelEnum(enum.Enum):
    """Tensor status."""
    BASE = "base"
    STATS = "stats"
    VALUE = 'value'


def get_ack_reply(state=0):
    """The the ack EventReply."""
    reply = EventReply()
    state_mapping = {
        0: EventReply.Status.OK,
        1: EventReply.Status.FAILED,
        2: EventReply.Status.PENDING
    }
    reply.status = state_mapping.get(state, EventReply.Status.FAILED)

    return reply


def wrap_reply_response(error_code=None, error_message=None):
    """
    Wrap reply response.

    Args:
        error_code (str): Error code. Default: None.
        error_message (str): Error message. Default: None.

    Returns:
        str, serialized response.
    """
    if error_code is None:
        reply = {'state': ReplyStates.SUCCESS.value}
    else:
        reply = {
            'state': ReplyStates.FAILED.value,
            'error_code': error_code,
            'error_message': error_message
        }

    return reply


def create_view_event_from_tensor_basic_info(tensors_info):
    """
    Create view event reply according to tensor names.

    Args:
        tensors_info (list[TensorBasicInfo]): The list of TensorBasicInfo. Each element has keys:
            `full_name`, `node_type`, `iter`.

    Returns:
        EventReply, the event reply with view cmd.
    """
    view_event = get_ack_reply()
    for tensor_info in tensors_info:
        node_type = tensor_info.node_type
        if node_type == NodeTypeEnum.CONST.value:
            truncate_tag = node_type == NodeTypeEnum.CONST.value
        else:
            truncate_tag = node_type == NodeTypeEnum.PARAMETER.value
        tensor_name = tensor_info.full_name
        # create view command
        ms_tensor = view_event.view_cmd.tensors.add()
        ms_tensor.node_name, ms_tensor.slot = tensor_name.rsplit(':', 1)
        ms_tensor.truncate = truncate_tag
        ms_tensor.iter = tensor_info.iter

    return view_event


def is_scope_type(node_type):
    """Judge whether the type is scope type."""
    return node_type.endswith('scope')


def is_cst_type(node_type):
    """Judge whether the type is const type."""
    return node_type == NodeTypeEnum.CONST.value


@enum.unique
class DebuggerServerMode(enum.Enum):
    """Debugger Server Mode."""
    ONLINE = 'online'
    OFFLINE = 'offline'


class DumpSettings(enum.Enum):
    """Dump settings."""
    E2E_DUMP_SETTINGS = 'e2e_dump_settings'
    COMMON_DUMP_SETTINGS = 'common_dump_settings'
    ASYNC_DUMP_SETTINGS = 'async_dump_settings'


def is_valid_rank_dir_name(name):
    """Check if the name followed the rank directory format."""
    return bool(re.search(r'^rank_\d+$', name))


def load_tensor(load_info, step, request_iterator, cache_store, rank_id):
    """Load tensor to tmp file."""
    file_mode = 0o600
    dir_mode = 0o700
    if load_info.get('prev') == 'prev':
        step -= 1
    tensor_stream = cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(rank_id)
    temp_dir = tempfile.TemporaryDirectory(dir=tensor_stream.download_mgr.temp_base_dir)
    os.chmod(temp_dir.name, dir_mode)
    node_name, slot = load_info.get('tensor_name').rsplit(':', 1)
    _, node_name = node_name.rsplit('/', 1)
    # Carry 2 digit for timestamp to ensure remaining 12 digit after round method
    file_name = get_download_file_name(load_info.get('node_type'), node_name, slot)
    file_path = os.path.join(temp_dir.name, file_name)
    tensor = next(request_iterator)
    header = _generate_npy_header(tensor)
    _write_tensor(file_path, header)
    os.chmod(file_path, file_mode)
    _write_tensor(file_path, tensor.tensor_content)
    for tensor in request_iterator:
        _write_tensor(file_path, tensor.tensor_content)
    add_to_download_mgr(file_name, file_path, temp_dir, load_info, step, rank_id, cache_store)


def add_to_download_mgr(file_name, file_path, temp_dir, load_info, step, rank_id, cache_store):
    """Add the tensor info to download manager"""
    tensor_info = {
        "tensor_name": load_info.get("tensor_name"),
        "graph_name": load_info.get("graph_name"),
        "step": step,
        "rank_id": rank_id
    }
    tensor_stream = cache_store.get_stream_handler(Streams.TENSOR).get_tensor_handler_by_rank_id(rank_id)
    tensor_stream.download_mgr.add(file_name, file_path, temp_dir, **tensor_info)
    metadata = cache_store.get_stream_handler(Streams.METADATA).get(['step', 'state'])
    ret = {
        'tensor_file': True,
        'node_name': load_info.get("node_name")
    }
    ret.update(metadata)
    cache_store.put_data(ret)


def get_download_file_name(node_type, node_name, slot):
    """Get file name."""
    timestamp_carry = 100
    file_name = "{}.{}.0.0.{}.output.{}.NONE.npy".format(node_type, node_name, round(time.time() * timestamp_carry),
                                                         slot)
    return file_name


def _generate_npy_header(tensor_proto):
    """Generate the header for npy file."""
    shape = tuple(tensor_proto.dims)
    if shape == (0,):
        shape = ()
    np_type = np.dtype(NUMPY_TYPE_MAP.get(DataType.Name(tensor_proto.data_type)))
    array_info = "{" + "'descr': {}, 'fortran_order': False, 'shape': {}".format(repr(np_type.str),
                                                                                 repr(shape)) + "}"
    array_info = array_info.encode('latin1')
    header_length = len(array_info) + 1
    # the length of npy header need to be divided by 64 exactly, so fill by b' '.
    padding_length = 64 - ((10 + header_length) % 64)
    header = b'\x93NUMPY' + bytes([1, 0]) + struct.pack('<H', header_length + padding_length) + \
             array_info + b' ' * padding_length + b'\n'
    return header


def _write_tensor(file_path, value):
    """Write into temp file."""
    flags = os.O_APPEND | os.O_CREAT
    # File owner has read and write access.
    modes = stat.S_IWUSR | stat.S_IRUSR
    with os.fdopen(os.open(file_path, flags, modes), 'ab') as fout:
        fout.write(value)


def convert_tensor_stats(tensor_data):
    """Convert tensor object to dict."""
    tensor_stats = Statistics({'is_bool': tensor_data.is_bool,
                               'max_value': tensor_data.max_value,
                               'min_value': tensor_data.min_value,
                               'avg_value': tensor_data.avg_value,
                               'count': tensor_data.count,
                               'neg_zero_count': tensor_data.neg_zero_count,
                               'pos_zero_count': tensor_data.pos_zero_count,
                               'zero_count': tensor_data.zero_count,
                               'nan_count': tensor_data.nan_count,
                               'neg_inf_count': tensor_data.neg_inf_count,
                               'pos_inf_count': tensor_data.pos_inf_count})
    return tensor_stats


def put_tensor_stats_in_cache(tensor_stats_list, tensor_protos, cur_step, tensor_stream):
    """Put tensor stats in cache."""
    update_data_flag = False
    for tensor_stats, tensor_proto in zip(tensor_stats_list, tensor_protos):
        name = ':'.join([tensor_proto.node_name, tensor_proto.slot])
        step = cur_step - 1 if tensor_proto.iter and cur_step > 0 else cur_step
        if tensor_stats is None or tensor_stats.get("tensor_base", {}).get("data_size", 0) == 0:
            has_update = tensor_stream.put_empty_tensor(name, step)
        else:
            has_update = tensor_stream.put_tensor_stats(name, step, tensor_stats.get("tensor_base"),
                                                        tensor_stats.get("tensor_stats"))
        if has_update:
            update_data_flag = True
    return update_data_flag


def put_tensor_base_in_cache(tensor_base_list, tensor_protos, cur_step, tensor_stream):
    """Put tensor base in cache."""
    update_data_flag = False
    for tensor_base, tensor_proto in zip(tensor_base_list, tensor_protos):
        name = ':'.join([tensor_proto.node_name, tensor_proto.slot])
        step = cur_step - 1 if tensor_proto.iter and cur_step > 0 else cur_step
        if tensor_base is None:
            has_update = tensor_stream.put_empty_tensor(name, step)
        else:
            has_update = tensor_stream.put_tensor_base(name, step, tensor_base)
        if has_update:
            update_data_flag = True
    return update_data_flag


def get_tensor_value(tensor_proto, tensor_contents, node_info, cur_step, oversize, data_size):
    """Get tensor value."""
    tensor_base = {
        'dtype': tensor_proto.data_type,
        'shape': tensor_proto.dims,
        'data_size': data_size
    }
    value = {
        'step': cur_step - 1 if tensor_proto.iter and cur_step > 0 else cur_step,
        'name': ':'.join([tensor_proto.node_name, tensor_proto.slot]),
        'tensor_base': tensor_base,
        'tensor_contents': tensor_contents if not oversize else [],
        'stats': bool(node_info and node_info.get('stats', False)),
        'oversize': oversize
    }
    return value


def validate_type(param, param_name, expect_type, expect_type_name=None):
    """Validate type."""
    expect_types = expect_type
    if isinstance(expect_types, list):
        expect_types = tuple(expect_types)
    if not isinstance(param, expect_types):
        expect_type_name = expect_type_name if expect_type_name else expect_type
        raise DebuggerParamTypeError(f"The type of {param_name} should be {expect_type_name}, but got {type(param)}.")


def parse_param_to_iterable_obj(param, param_name, expected_range=None, error_report=True):
    """
    Get iterable int objects which should in expected range.

    Args:
        param (Union[int, list[int], None], optional): The input param.
        param_name (str): The name of the param.
        expected_range (set[int]): The acceptable range of param value. Default None,
            no check for value.
        error_report (bool): Whether raise Exception if the value is not in expected_range.

    Returns:
        list[int], list of param objects.
    """
    if param is None:
        return expected_range
    if isinstance(param, set):
        param = list(param)
    if not isinstance(param, list):
        param = [param]
    # validate ranks
    res = []
    for param_item in param:
        validate_type(param_item, param_name, int, 'int, list[int] or None')
        if expected_range and param_item not in expected_range:
            if error_report:
                raise DebuggerParamValueError(f"The value should be in {expected_range}")
            continue
        res.append(param_item)
    return res


def validate_slots(slots):
    """Validate slots is List[int] or None."""
    if slots is None:
        return
    if not isinstance(slots, list):
        raise TypeError(f"The param `slots` only support list[int] or None, but got {type(slots)}")
    for slot in slots:
        validate_type(slot, 'slot', int, 'list[int] or None')


def gc_disable(func):
    """Wrapper for disable gc."""

    @wraps(func)
    def disable(*args, **kwargs):
        gc_threshold = gc.get_threshold()
        # disable gc collection during parse graphs
        gc.set_threshold(0)
        try:
            return func(*args, **kwargs)
        finally:
            gc.set_threshold(*gc_threshold)

    return disable


def enter_tag(func):
    """Wrapper for add enter and exit log."""

    @wraps(func)
    def add_log(*args, **kwargs):
        try:
            log.debug("Enter func: %s", func.__name__)
            return func(*args, **kwargs)
        finally:
            log.debug("Exit func: %s", func.__name__)

    return add_log



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the message handler."""
import uuid
from queue import Queue, Empty
from threading import Lock

from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase


class EventHandler(StreamHandlerBase):
    """Message Handler."""

    max_limit = 1000  # the max number of items in cache

    def __init__(self):
        self._prev_flag = str(uuid.uuid4())
        self._cur_flag = str(uuid.uuid4())
        self._next_idx = 0
        self._event_cache = [None] * self.max_limit
        self._pending_requests = {}
        self._lock = Lock()

    @property
    def next_pos(self):
        """The next pos to be updated in cache."""
        return ':'.join([self._cur_flag, str(self._next_idx)])

    def has_pos(self, pos):
        """Get the event according to pos."""
        cur_flag, cur_idx = self._parse_pos(pos)
        if cur_flag not in [self._cur_flag, self._prev_flag]:
            cur_flag, cur_idx = self._cur_flag, 0
        event = self._event_cache[cur_idx]
        if event is not None:
            if not cur_flag or (cur_flag == self._cur_flag and cur_idx < self._next_idx) or \
                    (cur_flag == self._prev_flag and cur_idx >= self._next_idx):
                return event

        return None

    def clean(self):
        """Clean event cache."""
        with self._lock:
            self._prev_flag = str(uuid.uuid4())
            self._cur_flag = str(uuid.uuid4())
            self._next_idx = 0
            self._event_cache = [None] * self.max_limit
            value = {'metadata': {'pos': '0'}}
            self.clean_pending_requests(value)
            log.debug("Clean event cache. %d request is waiting.", len(self._pending_requests))

    def put(self, value):
        """
        Put value into event_cache.

        Args:
            value (dict): The event to be put into cache.
        """
        if not isinstance(value, dict):
            log.error("Dict type required when put event message.")
            raise DebuggerParamValueError("Dict type required when put event message.")

        with self._lock:
            log.debug("Put the %d-th message into queue. \n %d requests is waiting.",
                      self._next_idx, len(self._pending_requests))
            cur_pos = self._next_idx
            # update next pos
            self._next_idx += 1
            if self._next_idx >= self.max_limit:
                self._next_idx = 0
                self._prev_flag = self._cur_flag
                self._cur_flag = str(uuid.uuid4())
            # set next pos
            if not value.get('metadata'):
                value['metadata'] = {}
            value['metadata']['pos'] = self.next_pos
            self._event_cache[cur_pos] = value
            # feed the value for pending requests
            self.clean_pending_requests(value)

    def clean_pending_requests(self, value):
        """Clean pending requests."""
        for _, request in self._pending_requests.items():
            request.put(value)
        self._pending_requests = {}

    def get(self, filter_condition=None):
        """
        Get the pos-th value from event_cache according to filter_condition.

        Args:
            filter_condition (str): The index of event in cache. Default: None.

        Returns:
            object, the pos-th event.
        """
        flag, idx = self._parse_pos(filter_condition)
        cur_id = str(uuid.uuid4())
        with self._lock:
            # reset the pos after the cache is re-initialized.
            if not flag or flag not in [self._cur_flag, self._prev_flag]:
                idx = 0
            # get event from cache immediately
            if idx != self._next_idx and self._event_cache[idx]:
                return self._event_cache[idx]
            # wait for the event
            cur_queue = Queue(maxsize=1)
            self._pending_requests[cur_id] = cur_queue
        # block until event has been received
        event = self._wait_for_event(cur_id, cur_queue, filter_condition)

        return event

    def _parse_pos(self, pos):
        """Get next pos according to input position."""
        elements = pos.split(':')
        try:
            idx = int(elements[-1])
        except ValueError:
            log.error("Invalid index. The index in pos should be digit but get pos:%s", pos)
            raise DebuggerParamValueError("Invalid pos.")

        if idx < 0 or idx >= self.max_limit:
            log.error("Invalid index. The index in pos should between [0, %d)", self.max_limit)
            raise DebuggerParamValueError(f"Invalid pos. {idx}")
        flag = elements[0] if len(elements) == 2 else ''

        return flag, idx

    def _wait_for_event(self, cur_id, cur_queue, pos):
        """Wait for the pos-th event."""
        try:
            # set the timeout to 25 seconds which is less the the timeout limit from UI
            event = cur_queue.get(timeout=25)
        except Empty:
            event = None

        if event is None:
            with self._lock:
                if self._pending_requests.get(cur_id):
                    self._pending_requests.pop(cur_id)
                log.debug("Clean timeout request. Left pending requests: %d",
                          len(self._pending_requests))
            event = {'metadata': {'pos': pos}}

        return event



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Profile api.

This module provides the interfaces to profile parallel strategy.
"""
import os

from flask import Blueprint, request, jsonify
from marshmallow import ValidationError

from mindinsight.conf import settings
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.exceptions.exceptions import ProfilerFileNotFoundException
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory
from mindinsight.profiler.common.util import check_train_job_and_profiler_dir
from mindinsight.profiler.analyser.parallel_strategy_analyser import Status, ParallelStrategyCache

BLUEPRINT = Blueprint("profile_parallel_strategy", __name__, url_prefix=settings.URL_PATH_PREFIX + settings.API_PREFIX)


def init_module(app):
    """
    Init module entry.

    Args:
        app: the application obj.

    """
    app.register_blueprint(BLUEPRINT)


@BLUEPRINT.route("/profile/parallel-strategy/graphs", methods=["GET"])
def get_parallel_strategy():
    """Get parallel strategy by train id."""
    train_id = request.args.get('train_id')
    stage_id = request.args.get('stage_id')
    if stage_id != "metadata":
        stage_id = int(stage_id) if stage_id else 'metadata'

    profiler_dir = os.path.realpath(os.path.join(settings.SUMMARY_BASE_DIR, train_id, 'profiler'))

    try:
        profiler_dir = validate_and_normalize_path(profiler_dir, 'profiler')
    except ValidationError as exc:
        raise ProfilerFileNotFoundException('Invalid cluster_profiler dir, detail: %s.' % str(exc))

    check_train_job_and_profiler_dir(profiler_dir)
    log.info("call get_parallel_strategy: %s, %s", str(train_id), str(stage_id))

    res_data = ParallelStrategyCache.get_cache(stage_id, profiler_dir)
    if res_data:
        return jsonify({'status': Status.FINISH.value, 'data': res_data})

    analyser = AnalyserFactory.instance().get_analyser('parallel_strategy', profiler_dir)
    analyser.load_by_stage_id(stage_id)
    res_data = analyser.get_data(stage_id)
    return jsonify(res_data)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Parse exceptions module."""

from mindinsight.utils.exceptions import MindInsightException
from mindinsight.utils.constant import GraphDomainErrors


class UnknownDataTypeError(MindInsightException):
    """Unknwn data type error."""

    def __init__(self, proto_type):
        super().__init__(
            error=GraphDomainErrors.UNKNOWN_DATA_TYPE_ERROR,
            message=str(proto_type))


class TupleGetitemIndexError(MindInsightException):
    """Tuple getitem index error."""

    def __init__(self, op_name, index_name):
        super().__init__(
            error=GraphDomainErrors.TUPLE_GETITEM_INDEX_ERROR,
            message=f'op : {op_name}, index: {index_name}')


class UnknownTensorError(MindInsightException):
    """Unknwn tensor error."""

    def __init__(self, is_op, file_name):
        super().__init__(
            error=GraphDomainErrors.UNKNOWN_TENSOR_ERROR,
            message=f'is_op : {is_op}, file_name: {file_name}')



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define custom exception."""

from mindinsight.utils.constant import DataVisualErrors
from mindinsight.utils.exceptions import MindInsightException


class RestfulApiNotExist(MindInsightException):
    """404 not found."""
    def __init__(self):
        error_msg = '404 Not Found.'
        super(RestfulApiNotExist, self).__init__(DataVisualErrors.RESTFUL_API_NOT_EXIST,
                                                 error_msg,
                                                 http_code=404)


class RequestMethodNotAllowed(MindInsightException):
    """Request method not allowed."""
    def __init__(self):
        error_msg = '405 Method Not Allowed.'
        super(RequestMethodNotAllowed, self).__init__(DataVisualErrors.REQUEST_METHOD_NOT_ALLOWED,
                                                      error_msg,
                                                      http_code=405)


class PathNotDirectoryError(MindInsightException):
    """Raised when specified path do not exist."""
    def __init__(self, error_detail):
        """Initialize PathNotExistError"""
        error_msg = 'Specified path is not a directory. Detail: {}'.format(error_detail)
        super(PathNotDirectoryError, self).__init__(DataVisualErrors.PATH_NOT_DIRECTORY_ERROR,
                                                    error_msg,
                                                    http_code=400)


class SummaryLogPathInvalid(MindInsightException):
    """No valid log file in the path."""
    def __init__(self):
        error_msg = 'No valid summary log file in path'
        super(SummaryLogPathInvalid, self).__init__(DataVisualErrors.SUMMARY_LOG_PATH_INVALID,
                                                    error_msg,
                                                    http_code=400)


class CRCLengthFailedError(MindInsightException):
    """CRC length fail, record corrupted."""
    def __init__(self, error_detail):
        error_msg = 'CRC Length Failed. Detail: %s' % error_detail
        super(CRCLengthFailedError, self).__init__(DataVisualErrors.CRC_LENGTH_FAILED,
                                                   error_msg,
                                                   http_code=400)


class SummaryLogIsLoading(MindInsightException):
    """Data is loading."""

    def __init__(self, error_detail):
        error_msg = "Data is loading. Detail: %s" % error_detail
        super(SummaryLogIsLoading, self).__init__(DataVisualErrors.SUMMARY_LOG_IS_LOADING,
                                                  error_msg,
                                                  http_code=400)


class NodeNotInGraphError(MindInsightException):
    """Can not find node in graph error."""
    def __init__(self, node_name, node_type=None):
        if node_type is not None:
            error_msg = f"Can not find node in graph by the given node name. node name: {node_name}, type: {node_type}."
        else:
            error_msg = f"Can not find node in graph by the given node name. node name: {node_name}."
        super(NodeNotInGraphError, self).__init__(DataVisualErrors.NODE_NOT_IN_GRAPH_ERROR,
                                                  error_msg,
                                                  http_code=400)


class MaxCountExceededError(MindInsightException):
    """Count is out of limit."""
    def __init__(self):
        error_msg = "Count is out of limit."
        super(MaxCountExceededError, self).__init__(DataVisualErrors.MAX_COUNT_EXCEEDED_ERROR,
                                                    error_msg,
                                                    http_code=400)


class TrainJobNotExistError(MindInsightException):
    """Can not find the given train job."""
    def __init__(self, error_detail=None):
        if error_detail is None:
            error_msg = f"Train job is not exist."
        else:
            error_msg = f"Train job is not exist. Detail: {error_detail}"
        super(TrainJobNotExistError, self).__init__(DataVisualErrors.TRAIN_JOB_NOT_EXIST,
                                                    error_msg,
                                                    http_code=400)


class QueryStringContainsNullByteError(MindInsightException):
    """Query string contains null byte error."""
    def __init__(self, error_detail):
        error_msg = f"Query string contains null byte error. Detail: {error_detail}"
        super(QueryStringContainsNullByteError, self).__init__(DataVisualErrors.QUERY_STRING_CONTAINS_NULL_BYTE,
                                                               error_msg,
                                                               http_code=400)


class PluginNotAvailableError(MindInsightException):
    """The given plugin is not available."""
    def __init__(self, error_detail):
        error_msg = f"Plugin is not available. Detail: {error_detail}"
        super(PluginNotAvailableError, self).__init__(DataVisualErrors.PLUGIN_NOT_AVAILABLE,
                                                      error_msg,
                                                      http_code=400)


class GraphNotExistError(MindInsightException):
    """Can not found the given graph."""
    def __init__(self, error_detail=None):
        error_msg = 'Graph is not exist.' if error_detail is None else f'Graph is not exist. Detail: {error_detail}'
        super(GraphNotExistError, self).__init__(DataVisualErrors.GRAPH_NOT_EXIST,
                                                 error_msg,
                                                 http_code=400)


class ImageNotExistError(MindInsightException):
    """Unable to get a image based on a given condition."""
    def __init__(self, error_detail):
        error_msg = f'Image is not exist. Detail: {error_detail}'
        super(ImageNotExistError, self).__init__(DataVisualErrors.IMAGE_NOT_EXIST,
                                                 error_msg,
                                                 http_code=400)


class ScalarNotExistError(MindInsightException):
    """Unable to get scalar values based on a given condition."""
    def __init__(self, error_detail):
        error_msg = f'Scalar value is not exist. Detail: {error_detail}'
        super(ScalarNotExistError, self).__init__(DataVisualErrors.SCALAR_NOT_EXIST,
                                                  error_msg,
                                                  http_code=400)


class HistogramNotExistError(MindInsightException):
    """Unable to get histogram values based on a given condition."""
    def __init__(self, error_detail):
        error_msg = f'Histogram value is not exist. Detail: {error_detail}'
        super(HistogramNotExistError, self).__init__(DataVisualErrors.HISTOGRAM_NOT_EXIST,
                                                     error_msg,
                                                     http_code=400)


class TensorNotExistError(MindInsightException):
    """Unable to get tensor values based on a given condition."""
    def __init__(self, error_detail):
        error_msg = f'Tensor value is not exist. Detail: {error_detail}'
        super(TensorNotExistError, self).__init__(DataVisualErrors.TENSOR_NOT_EXIST,
                                                  error_msg,
                                                  http_code=400)


class LandscapeNotExistError(MindInsightException):
    """Unable to get tensor values based on a given condition."""
    def __init__(self, error_detail=None):
        if error_detail is None:
            error_msg = "Landscape value does not exist."
        else:
            error_msg = f'Landscape value does not exist. Detail: {error_detail}'
        super(LandscapeNotExistError, self).__init__(DataVisualErrors.LANDSCAPE_NOT_EXIST,
                                                     error_msg,
                                                     http_code=400)


class StepTensorDataNotInCacheError(MindInsightException):
    """Tensor data with specific step does not in cache."""
    def __init__(self, error_detail):
        error_msg = f'Tensor data not in cache. Detail: {error_detail}'
        super(StepTensorDataNotInCacheError, self).__init__(DataVisualErrors.STEP_TENSOR_DATA_NOT_IN_CACHE,
                                                            error_msg,
                                                            http_code=400)


class ResponseDataExceedMaxValueError(MindInsightException):
    """Response data exceed max value based on a given condition."""
    def __init__(self, error_detail):
        error_msg = f'Response data exceed max value. Detail: {error_detail}'
        super(ResponseDataExceedMaxValueError, self).__init__(DataVisualErrors.MAX_RESPONSE_DATA_EXCEEDED_ERROR,
                                                              error_msg,
                                                              http_code=400)


class TrainJobDetailNotInCacheError(MindInsightException):
    """Detail info of given train job is not in cache."""
    def __init__(self, error_detail="no detail provided."):
        error_msg = f'Detail info of the given train job is not in cache. Detail: {error_detail}'
        super().__init__(DataVisualErrors.TRAIN_JOB_DETAIL_NOT_IN_CACHE,
                         error_msg,
                         http_code=400)


class TensorTooLargeError(MindInsightException):
    """The given tensor is too large to shown on UI."""
    def __init__(self, error_detail):
        error_msg = f'Tensor is too large to show on UI. Detail: {error_detail}'
        super(TensorTooLargeError, self).__init__(DataVisualErrors.TENSOR_TOO_LARGE,
                                                  error_msg,
                                                  http_code=400)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The parallel strategy graph classes."""
import copy
import sys
import time
from enum import Enum
from typing import List

from mindinsight.profiler.common.log import logger
from mindinsight.datavisual.common.enums import PluginNameEnum


class NodeType(Enum):
    """Define the types of nodes that require special processing."""
    PARAMETER = 'Parameter'
    CONST = 'Const'


class NodeInstanceType(Enum):
    """Define the actual role of the node."""
    # Means that the operator is used to do gradient aggregation
    GRADIENT_AGGREGATION = 'GradientAggregation'

    # Means that the operator is a redistribution insertion operator
    REDISTRIBUTION = 'Redistribution'


class DataType(Enum):
    """Define the data type, more detail refer the anf ir proto file."""
    DT_STRING = 'DT_STRING'
    DT_BOOL = 'DT_BOOL'
    DT_INT8 = 'DT_INT8'
    DT_INT16 = 'DT_INT16'
    DT_INT32 = 'DT_INT32'
    DT_INT64 = 'DT_INT64'
    DT_UINT8 = 'DT_UINT8'
    DT_UINT16 = 'DT_UINT16'
    DT_UINT32 = 'DT_UINT32'
    DT_UINT64 = 'DT_UINT64'
    DT_FLOAT16 = 'DT_FLOAT16'
    DT_FLOAT32 = 'DT_FLOAT32'
    DT_FLOAT64 = 'DT_FLOAT64'
    DT_TENSOR = 'DT_TENSOR'
    DT_LIST = 'DT_LIST'
    DT_TUPLE = 'DT_TUPLE'
    DT_NONE = 'DT_NONE'
    DT_TYPE = 'DT_TYPE'


class Node:
    """Define the attributes of a node."""

    def __init__(self, name, node_id, topo_index=-1):
        self._node_id = node_id
        self.name = name
        self.type = ""
        self.instance_type = ''
        self._attr = dict()
        self._input = []
        self._output = {}
        self.scope = ""
        self.output_shape = []
        self.output_data_type = ""

        # This value will be used as the priority field.
        self.topo_index = topo_index
        self.parallel_group = []
        self.parallel_strategy = []  # gen_strategy
        self.parallel_group_ranks = ''

    @property
    def node_id(self):
        """The id of this node, and id is unique in graph."""
        return self._node_id

    @property
    def attr(self):
        """Get node attr."""
        return self._attr

    def add_attr(self, attr_dict):
        """
        Update node attr.

        Args:
            attr_dict (dict[str, str]): The attr of node.
        """
        self._attr.update(attr_dict)

    @property
    def input(self):
        """
        Get all input of current node.

        Returns:
            list[str], refer to the input attr.
        """
        return self._input

    def add_input(self, name):
        """
        Update input.

        Args:
            name (str): The source node name.
        """
        self._input.append(name)

    def replace_input(self, inputs):
        """Replace input"""
        self._input = inputs

    def remove_input(self, name):
        """Remove the input from node."""
        self._input.remove(name)

    def to_dict(self):
        return dict(
            node_id=str(self.topo_index) if self.type != NodeType.CONST.value else self._node_id,
            name=self.name,
            scope=self.scope,
            type=self.type,
            instance_type=self.instance_type,
            attr=self._attr,
            input=self._input,
            output_shape=self.output_shape,
            output_data_type=self.output_data_type,
            parallel_group=self.parallel_group,
            parallel_shard=self.parallel_strategy,
            parallel_group_rank=self.parallel_group_ranks
        )


class Graph:
    """This class holds graph data parsed from files"""

    # Limit the size of a single attribute value per node to avoid storing too much data
    MAX_NODE_ATTRIBUTE_VALUE_BYTES = 1024

    def __init__(self):
        self.rank_id = 0
        self.op_nodes = {}
        self.const_nodes = {}
        self.parameter_nodes = {}

    def build_graph(self, graph_proto: dict, rank_id: str):
        start = time.time()
        self.rank_id = rank_id
        self._parse_graph_proto(graph_proto)
        logger.info("Parse graph proto success, op node count: %d, parameters count: %d, "
                    "const count: %d, use time: %s.",
                    len(self.op_nodes), len(self.parameter_nodes), len(self.const_nodes), (time.time() - start))

    def list_op_node_id(self):
        for node_id in self.op_nodes:
            yield node_id

    def get_op_node(self, node_id):
        return self.op_nodes[node_id]

    def _parse_graph_proto(self, graph_proto: dict):
        self._parse_const_nodes(graph_proto['constVals'])
        self._parse_parameter_nodes(graph_proto['parameters'])
        self._parse_op_nodes(graph_proto['node'])
        self._update_input()

    def _update_input(self):
        """Update op node input"""
        for op_node in self.op_nodes.values():
            inputs = list(op_node.input)
            new_inputs = []
            for input_name in inputs:
                if input_name in self.const_nodes:
                    new_inputs.append(input_name)
                    continue

                if input_name not in self.op_nodes:
                    op_node.remove_input(input_name)
                    continue
                src_node = self.get_op_node(input_name)
                new_inputs.append(str(src_node.topo_index))
            op_node.replace_input(new_inputs)

    def _parse_op_nodes(self, node_protos):
        """Transform the node proto into Node class."""
        index = 1
        for node_proto in node_protos:
            name = node_proto.get('name')
            full_name = node_proto.get('fullName')
            if not name or not full_name:
                logger.warning("Finding a node with an empty name or empty full_name will not save it.")
                continue
            scope = node_proto.get('scope', '')
            op_type = node_proto.get('opType', '')
            node_name = self._process_node_name(scope, op_type, name, full_name)
            node = Node(name=node_name, node_id=name, topo_index=index)
            index += 1
            node.scope = scope
            node.type = op_type

            attribute = node_proto.get('attribute', [])
            input_ = node_proto.get('input', [])
            self._parse_attr_proto(attribute, node)
            self._parse_input_proto(input_, node)
            if 'shapes' in node.attr:
                node.output_shape = node.attr['shapes']
            else:
                node.output_shape = self._get_shape_by_parse_type_proto(node_proto.get('outputType', dict()))
            node.output_data_type = self._get_data_type_by_parse_type_proto(node_proto.get('outputType', dict()))

            node.parallel_group = [node.attr['group']] if 'group' in node.attr else []
            node.parallel_strategy = node.attr.get('in_strategy', [])
            node.parallel_group_ranks = node.attr.get('group_ranks', '')

            instance_name = node_proto.get('instanceName', '')
            if 'grad_mirror' in instance_name:
                node.instance_type = NodeInstanceType.GRADIENT_AGGREGATION.value
            elif 'redistribution' in instance_name:
                node.instance_type = NodeInstanceType.REDISTRIBUTION.value
            else:
                node.instance_type = ''

            self._append_node(node)

    def _parse_const_nodes(self, consts):
        """
        Parse `anf_ir_pb2.NameValueProto` object, and create a const node.

        Args:
            consts (list[anf_ir_pb2.NameValueProto]): Refer to `anf_ir_pb2.NameValueProto` object.
        """
        for const in consts:
            key = const.get('key')
            if not key:
                logger.warning("Finding a const with an empty key will not save it.")
                continue

            node = Node(name=key, node_id=key)
            node.type = NodeType.CONST.value

            self._append_node(node)

    def _parse_parameter_nodes(self, parameter_protos: list):
        """
        Parse `anf_ir_pb2.ParameterProto` object, and create a const node.

        Args:
            parameter_protos (list[dict]): The dict object refers to anf_ir_pb2.ParameterProto.
        """
        for parameter in parameter_protos:
            name = parameter.get('name')
            if not name:
                logger.warning("Finding a parameter with an empty name will not be saved.")
                continue
            node = Node(name=name, node_id=name)
            # Note: Display parameter as Const
            node.type = NodeType.CONST.value
            self._append_node(node)

    def _append_node(self, node: Node):
        if node.type == NodeType.PARAMETER.value:
            self.parameter_nodes[node.node_id] = node
        elif node.type == NodeType.CONST.value:
            self.const_nodes[node.node_id] = node
        else:
            self.op_nodes[node.node_id] = node

    @staticmethod
    def _process_node_name(scope, node_type: str, node_id: str, full_name=None):
        """Handle the special node name."""
        if node_type == "Load":
            # The Load operator needs to be renamed as it has the same name with parameter
            node_name = f'{scope}/{node_type}-op{node_id}'
        elif not full_name or \
                any(full_name.lower().endswith(f'[:{plugin.value.lower()}]') for plugin in PluginNameEnum):
            # process summary node
            node_name = f'{scope}/{node_type}-op{node_id}'
        else:
            node_name = full_name

        return node_name

    def _parse_attr_proto(self, attributes, node):
        """
        Parse `anf_ir_pb2.AttributeProto` object., and Filters large attribute values.

        Args:
            attributes (list[anf_ir_pb2.AttributeProto]): Refer to `anf_ir_pb2.AttributeProto` object.
            node (Node): Refer to `Node` object, it is used to log message and update attr.
        """
        for attr in attributes:
            if sys.getsizeof(attr['value']) > self.MAX_NODE_ATTRIBUTE_VALUE_BYTES:
                message = f"The attribute value of node({node.name}) " \
                          f"is over {self.MAX_NODE_ATTRIBUTE_VALUE_BYTES} Bytes, will ignore."
                logger.warning(message)
                continue

            if attr['name'] == 'gen_strategy':
                # The gen_strategy value is equal in_strategy value, so we only need to show one strategy value in attr
                continue

            value = self._parse_value_proto(attr['value'])
            node.add_attr({attr['name']: str(value)})

    def _parse_value_proto(self, value_proto: dict):
        """Format the value proto into python base type."""
        actual_value = value_proto

        if value_proto['dtype'] == DataType.DT_STRING.value:
            actual_value = value_proto['strVal']
        elif value_proto['dtype'] in (DataType.DT_INT8.value, DataType.DT_INT16.value,
                                      DataType.DT_INT32.value, DataType.DT_INT64.value):
            actual_value = int(value_proto['intVal'])
        elif value_proto['dtype'] == DataType.DT_BOOL.value:
            actual_value = bool(int(value_proto['boolVal']))
        elif value_proto['dtype'] in (DataType.DT_UINT8.value, DataType.DT_UINT16.value,
                                      DataType.DT_UINT32.value, DataType.DT_UINT64.value):
            actual_value = int(value_proto['uintVal'])
        elif value_proto['dtype'] in (DataType.DT_FLOAT16.value, DataType.DT_FLOAT32.value, DataType.DT_FLOAT64.value):
            actual_value = float(value_proto['floatVal']) if value_proto['floatVal'] \
                else value_proto.get('doubleVal', 0)
        elif value_proto['dtype'] in (DataType.DT_TUPLE.value, DataType.DT_LIST.value):
            actual_value = []
            for value in value_proto.get('values', []):
                actual_value.append(self._parse_value_proto(value))
        elif value_proto['dtype'] == DataType.DT_NONE.value:
            actual_value = value_proto['strVal']
        elif value_proto['dtype'] == DataType.DT_TYPE.value:
            actual_value = value_proto.get('typeVal', {}).get('dataType', '')

        return actual_value

    @staticmethod
    def _parse_input_proto(input_protos: list, node: Node):
        """
        Parse `anf_ir_pb2.InputProto` object.

        Args:
            input_protos (list[dict]): The key of this param refer to `anf_ir_pb2.InputProto` object.
            node (Node): Refer to `Node` object, it is used to log message and update input.
        """
        for input_proto in input_protos:
            if not input_proto['name']:
                logger.warning("The name in input proto of node(%s) is empty, will ignore.", node.name)
                continue

            node.add_input(name=input_proto['name'])

    def _get_shape_by_parse_type_proto(self, type_proto: dict):
        """
        Parse proto's `message TypeProto` to get shape information.

        Args:
            type_proto (dict): The keys of this param refer to anf_ir_pb2.TypeProto.

        Returns:
            list, a list of shape.
        """
        shapes = []
        if not type_proto:
            return shapes

        if 'dataType' in type_proto:
            if type_proto['dataType'] != DataType.DT_TENSOR.value and \
                    type_proto['dataType'] != DataType.DT_TUPLE.value:
                return []
        if 'tensorType' in type_proto and 'shape' in type_proto['tensorType']:
            tensor_type = type_proto['tensorType']
            tensor_shape_proto = tensor_type['shape']
            shapes = [dim['size'] for dim in tensor_shape_proto['dim']]
        if 'sequenceType' in type_proto and 'elemTypes' in type_proto['sequenceType']:
            for elem_type in type_proto['sequenceType']['elemTypes']:
                shape = self._get_shape_by_parse_type_proto(elem_type)
                if shape:
                    shapes.append(shape)
        return shapes

    def _get_data_type_by_parse_type_proto(self, type_proto: dict):
        """
        Get data type by parse type proto object.

        The name of the DataType, refer to `anf_ir_pb2.DataType` object.
        If data type is tensor or tuple, the data name we return is `data_type[element_type, element_type]`.

        Args:
            type_proto (dict): The keys of this param refer to anf_ir_pb2.TypeProto.

        Returns:
            str, the data type.
        """
        if not type_proto:
            return ''

        data_type_name = type_proto['dataType']
        if type_proto['dataType'] == DataType.DT_TENSOR.value:
            tensor_type_proto = type_proto['tensorType']
            elem_type_name = tensor_type_proto['elemType']
            return f'{data_type_name}[{elem_type_name}]'

        if type_proto['dataType'] == DataType.DT_TUPLE.value:
            data_types = []
            for elem_type in type_proto['sequenceType']['elemTypes']:
                data_types.append(self._get_data_type_by_parse_type_proto(elem_type))
            return f'{data_type_name}{str(data_types)}'

        return data_type_name


class MergedGraph:
    """We will merge `Graph`s from the same stage into a graph, and call the new graph MergedGraph."""

    def __init__(self):
        self._rank_ids = []
        self._op_nodes = {}
        self._const_nodes = {}
        self._parameter_nodes = {}

    def merge_graphs(self, graphs: List[Graph]):
        """This function merge the same stage graphs into a graph."""
        if not graphs:
            logger.warning("Can not find any graph when merge graphs.")
            return

        for graph in graphs:
            self._rank_ids.append(graph.rank_id)

        # Graph has the same operator on the same stage, so we can use the node ids with any graphs.
        node_ids = graphs[0].list_op_node_id()
        for node_id in node_ids:
            node = graphs[0].get_op_node(node_id)
            new_node = copy.deepcopy(node)
            for graph in graphs[1:]:
                new_node.parallel_group.extend(graph.get_op_node(node_id).parallel_group)
                new_node.parallel_group = list(sorted(set(new_node.parallel_group)))

            self._op_nodes[new_node.node_id] = new_node

        self._const_nodes = graphs[0].const_nodes
        self._parameter_nodes = graphs[0].parameter_nodes

    def to_dict(self):
        result = dict(
            rank_ids=sorted(map(int, self._rank_ids)),
            op_nodes=[node.to_dict() for node in self._op_nodes.values()],
            const_nodes=[node.to_dict() for node in self._const_nodes.values()],
            parameter_nodes=[node.to_dict() for node in self._parameter_nodes.values()]
        )
        return result


class SupportedParallelType(Enum):
    DATA_PARALLEL = 'data_parallel'
    SEMI_AUTO_PARALLEL = 'semi_auto_parallel'
    AUTO_PARALLEL = 'auto_parallel'

    @classmethod
    def list_members(cls):
        """List all members."""
        return [member.value for member in cls]


class GraphManager:
    """
    This class is used to add original graphs and manage the merged graphs.

    We use this class to build the original computation graph and then merge the same stage graphs into MergedGraph.
    And we can query any merged graph by this class.

    Args:
        parallel_type (str): The parallel type, refer to `SupportedParallelType`.
        stage_devices (list): refer to MindSpore ccsrc/utils/profiling_parallel.proto
            ProfilingParallel.Config.stage_devices.

    Examples:
        >>> manager = GraphManager('AUTO_PARALLEL', parallel.config.stage_devices)
        >>> for file in sorted(path.rglob("parallel_strategy*.pb")):
        ...     with file.open(mode='rb') as fp:
        ...         manager.add_graph(parallel.graph, parallel.config.rank_id)
        >>> manager.merge_graph()
    """

    def __init__(self, parallel_type: str, stage_id: str, stage_devices: list):
        self._parallel_type = parallel_type
        self._stage_id = stage_id
        self._stage_devices = stage_devices
        self._merged_graph = MergedGraph()
        self._rank_graphs = {}

    def add_graph(self, graph: Graph, rank_id: str):
        """Add an original graph object which is parsed from pb files."""
        self._rank_graphs[rank_id] = graph

    def merge_graph(self):
        """Merge the same stage graphs into a merged graph."""
        graphs = []
        for rank_id in self._stage_devices:
            if rank_id not in self._rank_graphs:
                logger.warning("This rank id(%s) is not found in all the parsed files.", rank_id)
                continue
            graphs.append(self._rank_graphs[rank_id])

        if not graphs:
            logger.warning("There can not find any graph in stage %s.", self._stage_id)
            return

        self._merged_graph.merge_graphs(graphs)

    def to_dict(self):
        """Get all merged graphs and the training attributes."""
        if self._merged_graph:
            return self._merged_graph.to_dict()
        return None



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Static resource api."""
import os
import sys

from flask import current_app
from flask import make_response
from flask import redirect
from flask import Blueprint

from mindinsight.conf import settings


BLUEPRINT = Blueprint("static_resource", __name__, url_prefix=settings.URL_PATH_PREFIX)


@BLUEPRINT.route("/", methods=["GET"])
def index():
    """Interface to return index.html."""
    app_path = os.path.realpath(os.path.dirname(sys.argv[0]))
    index_html_file = os.path.realpath(os.path.join(app_path, current_app.static_folder, os.pardir, 'index.html'))
    with open(index_html_file, 'r') as file_pointer:
        html_content = file_pointer.read()
    return make_response(html_content)


@BLUEPRINT.route("/graphvizlib.wasm", methods=["GET"])
def return_wasm_file():
    """
    Interface to redirect graphvizlib.wasm

    When accessing the graphvizlib.wasm file in front module via Firefox browser, the file path will change to
    "/graphvizlib.wasm" which makes the computed diagram inaccessible. Redirecting the path to correct address can
    ensure the computed graph accessible properly.
    """
    return redirect(location="static/js/graphvizlib.wasm")


def init_module(app):
    """
    Init module entry.

    Args:
        app: the application obj.
    """
    app.register_blueprint(BLUEPRINT)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Log module."""

import sys
import os
import stat
import time
import fcntl
import logging
from logging.handlers import RotatingFileHandler

from mindinsight.conf import settings
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.utils.constant import GeneralErrors


class MultiCompatibleRotatingFileHandler(RotatingFileHandler):
    """Inherit RotatingFileHandler for multiprocess compatibility."""

    def rolling_rename(self):
        """Rolling rename log files."""
        for i in range(self.backupCount - 1, 0, -1):
            sfn = self.rotation_filename("%s.%d" % (self.baseFilename, i))
            dfn = self.rotation_filename("%s.%d" % (self.baseFilename, i + 1))
            if os.path.exists(sfn):
                if os.path.exists(dfn):
                    os.remove(dfn)
                os.chmod(sfn, stat.S_IRUSR)
                os.rename(sfn, dfn)

    def doRollover(self):
        """Do a rollover, as described in __init__()."""
        if self.stream:
            self.stream.close()
            self.stream = None

        flags = os.O_WRONLY | os.O_CREAT
        mode = stat.S_IRUSR | stat.S_IWUSR

        # Attain an exclusive lock with blocking mode by `fcntl` module.
        with os.fdopen(os.open(self.baseFilename, flags, mode), 'w') as fp:
            fcntl.lockf(fp.fileno(), fcntl.LOCK_EX)

        try:
            if self.backupCount > 0:
                self.rolling_rename()

            # dfn stands for destinated file name according to source codes in RotatingFileHandler.doRollover()
            dfn = self.rotation_filename(self.baseFilename + ".1")
            if os.path.exists(dfn):
                os.remove(dfn)

            os.chmod(self.baseFilename, stat.S_IRUSR)
            self.rotate(self.baseFilename, dfn)

            fp = os.fdopen(os.open(self.baseFilename, flags, mode), 'w')
            fp.close()

            if not self.delay:
                self.stream = self._open()

        except FileNotFoundError:
            # Suppress this exception for concurrency.
            pass

    def _open(self):
        """Open the current base file with the (original) mode and encoding."""
        new_log = super()._open()
        os.chmod(self.baseFilename, stat.S_IREAD | stat.S_IWRITE)
        return new_log


class MindInsightFormatter(logging.Formatter):
    """
    MindInsight formatter.
    """

    def __init__(self, sub_module, fmt=None, **kwargs):
        """
        Initialization of SlogFormatter.

        Args:
            sub_module (str): Sub module name, type is string.
            fmt (str): Specified format pattern, type is string.

        Returns:
            Formatter, instance of SlogFormatter.
        """
        super(MindInsightFormatter, self).__init__(fmt=fmt, **kwargs)
        self.sub_module = sub_module.upper()

    def formatTime(self, record, datefmt=None):
        """
        Overwrite for uniform format %Y-%m-%d-%H:%M:%S.SSS.SSS

        Args:
            record (LogRecord): Log record.
            datefmt (str): Date format, type is string.

        Returns:
            str, formatted timestamp, type is string.
        """
        created_time = self.converter(record.created)
        if datefmt:
            return time.strftime(datefmt, created_time)

        timestamp = time.strftime('%Y-%m-%d-%H:%M:%S', created_time)
        msecs = str(round(record.msecs * 1000)).zfill(6)
        return '{}.{}.{}'.format(timestamp, msecs[:3], msecs[3:])

    def formatMessage(self, record):
        """Escape the message before format."""
        record.message = ' '.join(record.message.split())
        return super().formatMessage(record)

    def format(self, record):
        """
        Apply log format with specified pattern.

        Args:
            record (str): Format pattern, type is string.

        Returns:
            str, formatted log content according to format pattern, type if string.
        """
        record.filepath = record.pathname[__file__.rfind('mindinsight'):]
        record.sub_module = self.sub_module
        return super().format(record)


def get_logger(sub_module, log_name):
    """
    Get logger by name and sub module.

    Args:
        sub_module (str): Sub module name, type is string.
        log_name (str): Log file name, type is  string.

    Returns:
        Logger, logger instance named by sub_module and log_name.
    """
    return logging.getLogger(name='{}.{}'.format(sub_module, log_name))


def setup_logger(sub_module, log_name, **kwargs):
    """
    Setup logger with sub module name and log file name.

    Args:
        sub_module (str): Sub module name, also for sub directory under logroot.
        log_name (str): Log name, also for log filename.
        console (bool): Whether to output log to stdout. Default: False.
        logfile (bool): Whether to output log to disk. Default: True.
        level (Enum): Log level. Default: INFO.
        formatter (str): Log format.
        propagate (bool): Whether to enable propagate feature. Default: False.
        maxBytes (int): Rotating max bytes. Default: 50M.
        backupCount (int): Rotating backup count. Default: 30.

    Returns:
        Logger, well-configured logger instance.

    Examples:
        >>> from mindinsight.utils.log import setup_logger
        >>> logger = setup_logger('datavisual', 'flask.request', level=logging.DEBUG)

        >>> from mindinsight.utils.log import get_logger
        >>> logger = get_logger('datavisual', 'flask.request')

        >>> import logging
        >>> logger = logging.getLogger('datavisual.flask.request')
    """

    if kwargs.get('sub_log_name', False):
        logger = get_logger(sub_module, kwargs['sub_log_name'])
    else:
        logger = get_logger(sub_module, log_name)
    if logger.hasHandlers():
        return logger

    level = kwargs.get('level', settings.LOG_LEVEL)
    formatter = kwargs.get('formatter', None)
    propagate = kwargs.get('propagate', False)

    logger.setLevel(level)
    logger.propagate = propagate

    if not formatter:
        formatter = settings.LOG_FORMAT

    if kwargs.get('console', False):
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.formatter = MindInsightFormatter(sub_module, formatter)
        logger.addHandler(console_handler)

    if kwargs.get('logfile', True):
        max_bytes = kwargs.get('maxBytes', settings.LOG_ROTATING_MAXBYTES)

        if not isinstance(max_bytes, int) or not max_bytes > 0:
            raise MindInsightException(GeneralErrors.PARAM_VALUE_ERROR,
                                       'maxBytes should be int type and > 0.')

        backup_count = kwargs.get('backupCount',
                                  settings.LOG_ROTATING_BACKUPCOUNT)

        if not isinstance(backup_count, int) or not backup_count > 0:
            raise MindInsightException(GeneralErrors.PARAM_VALUE_ERROR,
                                       'backupCount should be int type and > 0.')

        logfile_dir = os.path.join(settings.WORKSPACE, 'log', sub_module)
        os.makedirs(logfile_dir, mode=stat.S_IRWXU, exist_ok=True)

        logfile_handler = MultiCompatibleRotatingFileHandler(
            filename=os.path.join(logfile_dir, '{}.{}.log'.format(log_name, settings.PORT)),
            maxBytes=max_bytes,
            backupCount=backup_count,
            encoding='utf8'
        )
        logfile_handler.formatter = MindInsightFormatter(sub_module, formatter)
        logger.addHandler(logfile_handler)

    return logger



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""PB parser module."""
import numpy as np

from mindinsight.domain.graph.base import MindSporeType, InputType, OutputType
from mindinsight.domain.graph.base import NodeInput, NodeOutput, Tensor, Source, Constant, Parameter, Operator, Parser
from mindinsight.domain.graph.exceptions import UnknownDataTypeError, TupleGetitemIndexError
from mindinsight.domain.graph.proto import ms_graph_pb2 as graph_proto


class PBParser(Parser):
    """Protobuf file parser."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.proto = graph_proto
        self.dtype_mapping = {
            self.proto.DT_UNDEFINED: None,
            self.proto.DT_BOOL: 'bool',
            self.proto.DT_INT8: 'int8',
            self.proto.DT_INT16: 'int16',
            self.proto.DT_INT32: 'int32',
            self.proto.DT_INT64: 'int64',
            self.proto.DT_UINT8: 'uint8',
            self.proto.DT_UINT16: 'uint16',
            self.proto.DT_UINT32: 'uint32',
            self.proto.DT_UINT64: 'uint64',
            self.proto.DT_FLOAT16: 'float16',
            self.proto.DT_FLOAT32: 'float32',
            self.proto.DT_FLOAT64: 'float64',
            self.proto.DT_TENSOR: 'tensor',
            self.proto.DT_TUPLE: 'tuple',
            self.proto.DT_STRING: 'string',
            self.proto.DT_BASE_INT: MindSporeType.INT,
            self.proto.DT_BASE_UINT: MindSporeType.UINT,
            self.proto.DT_BASE_FLOAT: MindSporeType.FLOAT,
        }
        self.numpy_type_mapping = {
            self.proto.DT_BOOL: bool,

            self.proto.DT_INT8: np.int8,
            self.proto.DT_INT16: np.int16,
            self.proto.DT_INT32: np.int32,
            self.proto.DT_INT64: np.int64,

            self.proto.DT_FLOAT16: np.float16,
            self.proto.DT_FLOAT32: np.float32,
            self.proto.DT_FLOAT64: np.float64,

            self.proto.DT_STRING: str
        }
        self.int_types = (
            self.proto.DT_INT8,
            self.proto.DT_INT16,
            self.proto.DT_INT32,
            self.proto.DT_INT64,
        )
        self.uint_types = (
            self.proto.DT_UINT8,
            self.proto.DT_UINT16,
            self.proto.DT_UINT32,
            self.proto.DT_UINT64,
        )
        self.float_types = (
            self.proto.DT_FLOAT16,
            self.proto.DT_FLOAT32,
            self.proto.DT_FLOAT64,
        )

    def _parse_constants(self, pb_constant):
        """
        Parse constants.

        Args:
            pb_constant (Protobuf): Constant node.
        """
        constant = Constant(pb_constant)
        if not pb_constant.HasField('value'):
            constant.output = NodeOutput(OutputType.TENSOR)
            self.constants.append(constant)
            return

        if pb_constant.value.dtype in self.int_types:
            constant.output = NodeOutput(OutputType(self.dtype_mapping[pb_constant.value.dtype]))
            constant.output.info['value'] = pb_constant.value.int_val
            constant.output.info['np_value'] = np.array(pb_constant.value.int_val,
                                                        dtype=self.numpy_type_mapping[pb_constant.value.dtype])
        elif pb_constant.value.dtype in self.float_types:
            constant.output = NodeOutput(OutputType(self.dtype_mapping[pb_constant.value.dtype]))
            constant.output.info['value'] = pb_constant.value.float_val
            constant.output.info['np_value'] = np.array(pb_constant.value.float_val,
                                                        dtype=self.numpy_type_mapping[pb_constant.value.dtype])
        elif pb_constant.value.dtype == self.proto.DT_BOOL:
            constant.output = NodeOutput(OutputType(self.dtype_mapping[pb_constant.value.dtype]))
            constant.output.info['value'] = pb_constant.value.bool_val
            constant.output.info['np_value'] = np.array(pb_constant.value.bool_val,
                                                        dtype=self.numpy_type_mapping[pb_constant.value.dtype])
        elif pb_constant.value.dtype == self.proto.DT_STRING:
            constant.output = NodeOutput(OutputType(self.dtype_mapping[pb_constant.value.dtype]))
            constant.output.info['value'] = pb_constant.value.str_val
            constant.output.info['np_value'] = np.array(pb_constant.value.str_val,
                                                        dtype=self.numpy_type_mapping[pb_constant.value.dtype])
        elif pb_constant.value.dtype == self.proto.DT_TENSOR:
            constant.output = NodeOutput(OutputType.TENSOR)
            if pb_constant.value.tensor_val.data_type:
                constant.output.info['dtype'] = self.dtype_mapping[pb_constant.value.tensor_val.data_type]
                constant.output.info['shape'] = tuple(pb_constant.value.tensor_val.dims)

            feature = Tensor.FEATURE(type='', id=constant.name, io='output')
            values = self.tensor_mapping.get(feature, [])
            if len(values) == 1:
                value = values[0]
                constant.output.info['tensor'] = Tensor(constant.name, 0, value.path)
        else:
            constant.output = NodeOutput(OutputType.NONE)

        self.constants.append(constant)

    def _parse_parameters(self, pb_parameter):
        """
        Parse parameters.

        Args:
            pb_parameter (Protobuf): Parameter node.
        """
        parameter = Parameter(pb_parameter)
        output = NodeOutput(OutputType.TENSOR)
        if pb_parameter.type.tensor_type.elem_type == self.proto.DT_UNDEFINED:
            output.type = OutputType.NONE
        else:
            output.info['dtype'] = self.dtype_mapping[pb_parameter.type.tensor_type.elem_type]
            output.info['shape'] = tuple([dim.size for dim in pb_parameter.type.tensor_type.shape.dim])

            feature = Tensor.FEATURE(type='', id=parameter.name, io='output')
            values = self.tensor_mapping.get(feature, [])
            if len(values) == 1:
                value = values[0]
                output.info['tensor'] = Tensor(parameter.name, 0, value.path)

        parameter.output = output
        self.parameters.append(parameter)

    def _get_proto_value(self, value):
        """
        Get proto value.

        Args:
            value (Protobuf): Protobuf value.

        Returns:
            any, proto value.

        Raises:
            UnknownDataTypeError: If data type of protobuf value can not be recognized.
        """
        if value.dtype == self.proto.DT_UNDEFINED:
            return None
        if value.dtype == self.proto.DT_TYPE:
            return self.dtype_mapping[value.type_val.data_type]
        if value.dtype == self.proto.DT_BOOL:
            return value.bool_val
        if value.dtype == self.proto.DT_STRING:
            return value.str_val
        if value.dtype in self.uint_types:
            return value.uint_val
        if value.dtype in self.int_types:
            return value.int_val
        if value.dtype in self.float_types:
            return value.float_val
        if value.dtype in (self.proto.DT_LIST, self.proto.DT_TUPLE):
            value_items = []
            for value_item in value.values:
                value_items.append(self._get_proto_value(value_item))
            if value.dtype == self.proto.DT_TUPLE:
                value_items = tuple(value_items)
            return value_items
        raise UnknownDataTypeError(value.dtype)

    def _find_operator_output_tensors(self, operator):
        """
        Find operator output tensors.

        Args:
            operator (Operator): Operator object.
        """
        output = operator.output
        if operator.type in ('tuple_getitem', 'make_tuple') or output.type in (OutputType.BOOL, OutputType.NONE):
            return

        is_load_op = False
        if operator.full_name.find('-op') == -1:
            is_load_op = True
            feature = Tensor.FEATURE(type='', id=operator.full_name, io='output')
        else:
            _, op_id = operator.full_name.split('-op')
            feature = Tensor.FEATURE(type=operator.type, id=op_id, io='output')

        values = self.tensor_mapping.get(feature)
        if not values:
            return

        if output.type == OutputType.TENSOR and len(values) == 1:
            value = values[0]
            if is_load_op:
                output.info['tensor'] = Tensor(operator.full_name, 0, value.path)
            else:
                output.info['tensor'] = Tensor(op_id, value.index, value.path)
        elif output.type == OutputType.TUPLE and len(values) == len(output.info['dtypes']):
            for value in values:
                output.info['tensors'][value.index] = Tensor(op_id, value.index, value.path)

    def _process_operator_input(self, operator, input_index, input_types):
        """
        Process operator input.

        Args:
            operator (Operator): Operator.
            input_index (int): Input index.
            input_types (dict): Input types.
        """
        op_input = operator.inputs[input_index]
        node = input_types[op_input.type].get(op_input.name)
        if not node:
            return
        node.downstream.append(operator.op_id)
        if op_input.type == InputType.OPERATOR:
            op_input.op_id = node.op_id

        op_input.info = {}

        if op_input.type in (InputType.PARAMETER, InputType.CONSTANT):
            op_input.info = node.output.info.copy() if node.output.info else None
            return

        if operator.type in ('tuple_getitem', 'TupleGetItem'):
            op_input.info['dtype'] = OutputType.TUPLE
            return

        if node.output.type == OutputType.TENSOR:
            op_input.info = node.output.info.copy()
        else:
            op_input.info['dtype'] = node.output.type

        if node.full_name.find('-op') == -1:
            op_id = node.full_name
            feature = Tensor.FEATURE(type='', id=node.full_name, io='output')
        else:
            _, op_id = node.full_name.split('-op')
            feature = Tensor.FEATURE(type=node.type, id=op_id, io='input')

        values = self.tensor_mapping.get(feature)
        if values and len(values) == len(operator.inputs) and op_input.info['tensor'] is None:
            value = values[input_index]
            op_input.info['tensor'] = Tensor(op_id, value.index, value.path)

    def _parse_operators(self, pb_operator):
        """
        Parse operator.

        Args:
            pb_operator (Protobuf): Operator node.

        Raises:
            UnknownDataTypeError: If data type of protobuf value can not be recognized.
        """
        if pb_operator.name == '':
            return

        operator = Operator(pb_operator)
        self.operators.append(operator)

        # parse source code
        if getattr(pb_operator, 'source_address', None):
            operator.stack = Source.build_stack_from_source_address(pb_operator.source_address)

        # parse attrs
        for attr in pb_operator.attribute:
            operator.attrs[attr.name] = self._get_proto_value(attr.value)
            if attr.name in ('input_names', 'output_names'):
                operator.attrs[attr.name] = list(operator.attrs[attr.name])

        self._parse_operator_inputs(operator, pb_operator)
        self._parse_operator_outputs(operator, pb_operator)
        self._find_operator_output_tensors(operator)

    def _parse_operator_outputs(self, operator, pb_operator):
        """Parse the outputs of the operator."""
        proto_output = pb_operator.output_type
        if proto_output.data_type in (self.proto.DT_UNDEFINED, self.proto.DT_NONE):
            output = NodeOutput(OutputType.NONE)
        elif proto_output.data_type == self.proto.DT_BOOL:
            output = NodeOutput(OutputType.BOOL)
        elif proto_output.data_type == self.proto.DT_TENSOR:
            output = NodeOutput(OutputType.TENSOR)
            output.info['dtype'] = self.dtype_mapping[proto_output.tensor_type.elem_type]
            if proto_output.tensor_type.shape:
                output.info['shape'] = tuple([dim.size for dim in proto_output.tensor_type.shape.dim])
        elif proto_output.data_type == self.proto.DT_TUPLE:
            output = NodeOutput(OutputType.TUPLE)
            for elem_type in proto_output.sequence_type.elem_types:
                dtype = self._get_tuple_item_dtype(elem_type)
                if dtype is None:
                    continue
                output.info['dtypes'].append(dtype)
                output.info['shapes'].append(None)
                output.info['tensors'].append(None)
                output.slot_size += 1
        else:
            raise UnknownDataTypeError(proto_output.data_type)
        operator.output = output

    def _parse_operator_inputs(self, operator, pb_operator):
        """Parse the inputs of the operator."""
        cst_mapping = dict((cst.name, cst) for cst in self.constants)
        param_mapping = dict((param.name, param) for param in self.parameters)
        op_mapping = dict((op.name, op) for op in self.operators)
        for pb_input in pb_operator.input:
            input_type = InputType.REFERENCE
            if pb_input.name in cst_mapping:
                input_type = InputType.CONSTANT
            elif pb_input.name in param_mapping:
                input_type = InputType.PARAMETER
            elif pb_input.name in op_mapping:
                input_type = InputType.OPERATOR

            op_input = NodeInput(input_type, pb_input.name)
            if input_type == InputType.OPERATOR:
                op_input.op_id = op_mapping[op_input.name].op_id

            operator.inputs.append(op_input)

    def _get_tuple_item_dtype(self, elem_type):
        """
        Get tuple item dtype.

        Args:
            elem_type (TypeProto) : TypeProto of tuple item operator.

        Returns:
            any, tuple item dtype.
        """
        if elem_type.tensor_type.elem_type != self.proto.DT_UNDEFINED:
            return self.dtype_mapping[elem_type.tensor_type.elem_type]

        return self.dtype_mapping[elem_type.data_type]

    def _get_tuple_getitem_index(self, operator, constant_mapping):
        """
        Get tuple_getitem index.

        Args:
            operator (Operator) : Operator.
            constant_mapping (dict) : Constant mapping.

        Returns:
            int, tuple_getitem index.

        Raises:
            TupleGetitemIndexError: If tuple_getitem index error occurs.
        """
        if operator.inputs[1].type == InputType.CONSTANT:
            constant = constant_mapping[operator.inputs[1].name]
            if constant.output.type in NodeOutput.SCALAR_TYPES:
                return int(constant.output.info['value'])
            raise TupleGetitemIndexError(operator.name, constant.name)

        if operator.inputs[1].type == InputType.SCALAR:
            return int(operator.inputs[1].name)
        raise TupleGetitemIndexError(operator.name, f'{operator.inputs[1].name}')

    def _process_tuple_operator(self, operator, operator_mapping, constant_mapping):
        """
        Process tuple operator.

        Args:
            operator (Operator): Tuple operator.
            operator_mapping (dict): Dict mapping of operators.
            constant_mapping (dict): Dict mapping of constants.
        """
        if operator.type in ('make_tuple', 'MakeTuple'):
            return

        if operator.type in ('tuple_getitem', 'TupleGetItem'):
            index = self._get_tuple_getitem_index(operator, constant_mapping)
            node = operator_mapping.get(operator.inputs[0].name)
            if node is None:
                return
            output = node.output
            if output.type == OutputType.TUPLE and len(output.info['tensors']) > index:
                operator.output.info['tensor'] = output.info['tensors'][index]

    def _process_transition_operator(self, operator, input_types):
        """
        Process transition operator.

        Args:
            operator (Operator): Transition operator.
            input_types (dict): Dict mapping of input types.
        """
        if operator.output.info['tensor'] is None:
            op_input = operator.inputs[0]
            if op_input.type in input_types:
                node = input_types[op_input.type][op_input.name]
                operator.output.info['tensor'] = node.output.info['tensor']

    def _post_process(self):
        """Post-process."""
        constant_mapping = dict((constant.name, constant) for constant in self.constants)
        parameter_mapping = dict((parameter.name, parameter) for parameter in self.parameters)
        operator_mapping = dict((operator.name, operator) for operator in self.operators)
        input_types = {
            InputType.CONSTANT: constant_mapping,
            InputType.PARAMETER: parameter_mapping,
            InputType.OPERATOR: operator_mapping,
        }

        tuple_operator_types = ('make_tuple', 'MakeTuple', 'tuple_getitem', 'TupleGetItem')
        transition_operator_types = ('Squeeze', 'Reshape', 'ExpandDims', 'Flatten')

        for operator in self.operators:
            if operator.type in tuple_operator_types:
                self._process_tuple_operator(operator, operator_mapping, constant_mapping)

            elif operator.type in transition_operator_types:
                self._process_transition_operator(operator, input_types)

            elif operator.type == 'Depend':
                if operator.output.type in (OutputType.NONE, OutputType.BOOL) \
                        or (operator.output.type == OutputType.TENSOR and operator.output.info['tensor'] is None):
                    op_input = operator.inputs[0]
                    if op_input.type == InputType.OPERATOR:
                        node = operator_mapping.get(op_input.name)
                        if node is None:
                            continue
                        operator.output = node.output

            elif operator.type == 'Assign' and len(operator.inputs) == 3:
                operator.inputs = operator.inputs[1:]

            for input_index, op_input in enumerate(operator.inputs):
                if op_input.type in input_types:
                    self._process_operator_input(operator, input_index, input_types)

    def parse(self):
        """Parse."""
        self.tensor_mapping = Tensor.scan_tensors(self.tensor_dir)

        # parse constants
        for pb_constant in self.graph_data.const_vals:
            self._parse_constants(pb_constant)

        # parse parameters:
        for pb_parameter in self.graph_data.parameters:
            self._parse_parameters(pb_parameter)

        # parse operators
        for pb_operator in self.graph_data.node:
            self._parse_operators(pb_operator)

        # post-process
        self._post_process()
